from pathlib import Path

import pytest
from affine import Affine

from odc.geo.gcp import GCPGeoBox
from odc.geo.geobox import GeoBox

rioxarray = pytest.importorskip("rioxarray")


@pytest.mark.parametrize("fname", ["au-gcp.tif", "au-3577.tif", "au-3577-rotated.tif"])
@pytest.mark.parametrize("parse_coordinates", [True, False])
def test_rioxarray_interop(data_dir: Path, fname: str, parse_coordinates: bool):
    xx = rioxarray.open_rasterio(
        data_dir / fname, chunks=64, parse_coordinates=parse_coordinates
    )
    assert xx.rio.crs is not None
    assert xx.rio.crs == xx.odc.crs

    if "-gcp" in fname:
        assert isinstance(xx.odc.geobox, GCPGeoBox)
    else:
        assert isinstance(xx.rio.transform(), Affine)
        assert xx.rio.transform() == xx.odc.transform
        assert isinstance(xx.odc.geobox, GeoBox)
        assert xx[0, 1:2, -1:].odc.geobox is not None



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2020 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
"""
Add ``.odc.`` extension to :py:class:`xarray.Dataset` and :class:`xarray.DataArray`.
"""

from __future__ import annotations

import functools
import json
import math
import warnings
from dataclasses import dataclass
from datetime import datetime
from typing import (
    Any,
    Callable,
    Dict,
    Hashable,
    List,
    Literal,
    Optional,
    Set,
    Tuple,
    TypeVar,
    Union,
)

import numpy
import xarray
from affine import Affine

from ._interop import have, is_dask_collection
from ._rgba import colorize, to_rgba
from .crs import CRS, CRSError, SomeCRS, norm_crs_or_error
from .gcp import GCPGeoBox, GCPMapping
from .geobox import Coordinate, GeoBox, GeoboxAnchor
from .geom import Geometry
from .math import (
    affine_from_axis,
    approx_equal_affine,
    is_affine_st,
    is_nodata_empty,
    maybe_int,
    resolution_from_affine,
    resolve_fill_value,
    resolve_nodata,
)
from .overlap import compute_output_geobox
from .roi import roi_is_empty
from .types import Nodata, Resolution, SomeNodata, SomeResolution, SomeShape, Unset, xy_

# pylint: disable=import-outside-toplevel
# pylint: disable=too-many-lines
if have.rasterio:
    from ._compress import compress
    from ._map import add_to, explore
    from .cog import to_cog, write_cog
    from .warp import rio_reproject

XarrayObject = Union[xarray.DataArray, xarray.Dataset]
XrT = TypeVar("XrT", xarray.DataArray, xarray.Dataset)
F = TypeVar("F", bound=Callable)
SomeGeoBox = Union[GeoBox, GCPGeoBox]

_DEFAULT_CRS_COORD_NAME = "spatial_ref"

# these attributes are pruned during reproject
SPATIAL_ATTRIBUTES = ("crs", "crs_wkt", "grid_mapping", "gcps", "epsg")
NODATA_ATTRIBUTES = ("nodata", "_FillValue")
REPROJECT_SKIP_ATTRS: set[str] = set(SPATIAL_ATTRIBUTES + NODATA_ATTRIBUTES)

# dimensions with these names are considered spatial
STANDARD_SPATIAL_DIMS = [
    ("y", "x"),
    ("yc", "xc"),
    ("latitude", "longitude"),
    ("lat", "lon"),
]

_NoValue = Unset()


@dataclass
class GeoState:
    """
    Geospatial information for xarray object.
    """

    spatial_dims: Optional[Tuple[str, str]] = None
    crs_coord: Optional[xarray.DataArray] = None
    transform: Optional[Affine] = None
    crs: Optional[CRS] = None
    geobox: Optional[SomeGeoBox] = None
    gcp: Optional[GCPMapping] = None


def _get_crs_from_attrs(obj: XarrayObject, sdims: Tuple[str, str]) -> Optional[CRS]:
    """
    Looks for attribute named ``crs`` containing CRS string.

    - Checks spatials coords attrs
    - Checks data variable attrs
    - Checks dataset attrs

    Returns
    =======
    Content for `.attrs[crs]` usually it's a string
    None if not present in any of the places listed above
    """
    crs_set: Set[CRS] = set()

    def _add_candidate(crs):
        if crs is None:
            return
        if isinstance(crs, str):
            try:
                crs_set.add(CRS(crs))
            except CRSError:
                warnings.warn(f"Failed to parse CRS: {crs}")
        elif isinstance(crs, CRS):
            # support current bad behaviour of injecting CRS directly into
            # attributes in example notebooks
            crs_set.add(crs)
        else:
            warnings.warn(f"Ignoring crs attribute of type: {type(crs)}")

    def process_attrs(attrs):
        _add_candidate(attrs.get("crs", None))
        _add_candidate(attrs.get("crs_wkt", None))

    def process_datavar(x):
        process_attrs(x.attrs)
        for dim in sdims:
            if dim in x.coords:
                process_attrs(x.coords[dim].attrs)

    if isinstance(obj, xarray.Dataset):
        process_attrs(obj.attrs)
        for dv in obj.data_vars.values():
            process_datavar(dv)
    else:
        process_datavar(obj)

    crs = None
    if len(crs_set) >= 1:
        crs = crs_set.pop()

    if len(crs_set) > 0:
        if any(other != crs for other in crs_set):
            warnings.warn("Have several candidates for a CRS")

    return crs


def spatial_dims(
    xx: Union[xarray.DataArray, xarray.Dataset], relaxed: bool = False
) -> Optional[Tuple[str, str]]:
    """
    Find spatial dimensions of ``xx``.

    Checks for presence of dimensions named:
    ``y, x | latitude, longitude | lat, lon``

    If ``relaxed=True`` and none of the above dimension names are found,
    assume that last two dimensions are spatial dimensions.

    :returns: ``None`` if no dimensions with expected names are found
    :returns: ``('y', 'x') | ('latitude', 'longitude') | ('lat', 'lon')``
    """

    def skip_dim(dim: str) -> bool:
        if dim in ("time", "band", "bands", "wavelength", "wavelengths"):
            return True

        # skip dimensions without coord of the same name
        if dim not in xx.coords:
            return True

        coord = xx.coords[dim]
        # Primary coordinate for spatial dimension must have floating point type
        if coord.dtype.kind != "f":
            return True

        return False

    _dims = [str(dim) for dim in xx.dims]
    dims = set(_dims)
    for guess in STANDARD_SPATIAL_DIMS:
        if dims.issuperset(guess):
            return guess

    _dims = [dim for dim in _dims if not skip_dim(str(dim))]

    if relaxed and len(_dims) >= 2:
        return _dims[-2], _dims[-1]

    return None


def _mk_crs_coord(
    crs: CRS,
    name: str = _DEFAULT_CRS_COORD_NAME,
    gcps=None,
    transform: Optional[Affine] = None,
) -> xarray.DataArray:
    # pylint: disable=protected-access

    cf = crs.proj.to_cf()
    epsg = 0 if crs.epsg is None else crs.epsg
    crs_wkt = cf.get("crs_wkt", None) or crs.wkt

    if gcps is not None:
        # Store as string
        cf["gcps"] = json.dumps(_gcps_to_json(gcps))

    if transform is not None:
        cf["GeoTransform"] = _render_geo_transform(transform, precision=24)

    return xarray.DataArray(
        numpy.asarray(epsg, "int32"),
        name=name,
        dims=(),
        attrs={"spatial_ref": crs_wkt, **cf},
    )


def _gcps_to_json(gcps):
    def _to_feature(p):
        coords = [p.x, p.y] if p.z is None else [p.x, p.y, p.z]

        return {
            "type": "Feature",
            "properties": {
                "id": str(p.id),
                "info": (p.info or ""),
                "row": p.row,
                "col": p.col,
            },
            "geometry": {"type": "Point", "coordinates": coords},
        }

    return {"type": "FeatureCollection", "features": list(map(_to_feature, gcps))}


def _coord_to_xr(name: str, c: Coordinate, **attrs) -> xarray.DataArray:
    """
    Construct xr.DataArray from named Coordinate object.

    This can then be used to define coordinates for ``xr.Dataset|xr.DataArray``
    """
    attrs = {"units": c.units, "resolution": c.resolution, **attrs}
    return xarray.DataArray(
        c.values, coords={name: c.values}, dims=(name,), attrs=attrs
    )


def assign_crs(
    xx: XrT,
    crs: SomeCRS,
    crs_coord_name: str = _DEFAULT_CRS_COORD_NAME,
) -> XrT:
    """
    Assign CRS for a non-georegistered array or dataset.

    Returns a new object with CRS information populated.

    .. code-block:: python

        xx = xr.open_rasterio("some-file.tif")
        print(xx.odc.crs)
        print(xx.astype("float32").crs)


    :param xx: :py:class:`~xarray.Dataset` or :py:class:`~xarray.DataArray`
    :param crs: CRS to assign
    :param crs_coord_name: how to name crs coordinate (defaults to ``spatial_ref``)
    """
    crs = norm_crs_or_error(crs)
    crs_coord = _mk_crs_coord(crs, name=crs_coord_name)
    xx = xx.assign_coords({crs_coord_name: crs_coord})

    if isinstance(xx, xarray.DataArray):
        xx.encoding.update(grid_mapping=crs_coord_name)
    elif isinstance(xx, xarray.Dataset):
        for band in xx.data_vars.values():
            band.encoding.update(grid_mapping=crs_coord_name)

    return xx


def mask(
    xx: XrT, poly: Geometry, invert: bool = False, all_touched: bool = True
) -> XrT:
    """
    Apply a polygon geometry as a mask, setting all
    :py:class:`xarray.Dataset` or :py:class:`xarray.DataArray` pixels
    outside the rasterized polygon to ``NaN``.

    :param xx:
       :py:class:`~xarray.Dataset` or :py:class:`~xarray.DataArray`.

    :param poly:
       A :py:class:`odc.geo.geom.Geometry` polygon used to mask ``xx``.

    :param invert:
        Whether to invert the mask before applying it to ``xx``. If
        ``True``, only pixels inside of ``poly`` will be masked.

    :param all_touched:
        If ``True``, the rasterize step will burn in all pixels touched
        by ``poly``. If ``False``, only pixels whose centers are within
        the polygon or that are selected by Bresenham's line algorithm
        will be burned in.

    :return:
        A :py:class:`~xarray.Dataset` or :py:class:`~xarray.DataArray`
        masked by ``poly``.

    .. seealso:: :py:meth:`odc.geo.xr.rasterize`
    """
    # Rasterise `poly` into geobox of `xx`
    rasterized = rasterize(
        poly=poly,
        how=xx.odc.geobox,
        all_touched=all_touched,
        value_inside=not invert,
    )

    # Mask data outside rasterized `poly`
    xx_masked = xx.where(rasterized.data)

    # Remove nodata attribute from arrays
    if isinstance(xx_masked, xarray.Dataset):
        for var in xx_masked.data_vars:
            xx_masked[var].attrs.pop("nodata", None)
    else:
        xx_masked.attrs.pop("nodata", None)

    return xx_masked


def crop(
    xx: XrT, poly: Geometry, apply_mask: bool = True, all_touched: bool = True
) -> XrT:
    """
    Crops and optionally mask an :py:class:`xarray.Dataset` or
    :py:class:`xarray.DataArray` to the spatial extent of a geometry.

    :param xx:
       :py:class:`~xarray.Dataset` or :py:class:`~xarray.DataArray`.

    :param poly:
       A :py:class:`odc.geo.geom.Geometry` polygon used to crop ``xx``.

    :param apply_mask:
       Whether to mask out pixels outside of the rasterized extent of
       ``poly`` by setting them to ``NaN``.

    :param all_touched:
        If ``True`` and ``apply_mask=True``, the rasterize step will
        burn in all pixels touched by ``poly``. If ``False``, only
        pixels whose centers are within the polygon or that are selected
        by Bresenham's line algorithm will be burned in.

    :return:
        A :py:class:`~xarray.Dataset` or :py:class:`~xarray.DataArray`
        cropped and optionally masked to the spatial extent of ``poly``.

    .. seealso:: :py:meth:`odc.geo.xr.mask`
    """
    meta: ODCExtension = xx.odc
    sdims = meta.spatial_dims
    gbox = meta.geobox

    if sdims is None or gbox is None:
        raise ValueError("Can't locate spatial dimensions")

    if not isinstance(gbox, GeoBox):
        raise ValueError("Can't crop GCPGeoBox")

    # Create new geobox with pixel grid of `xx` but enclosing `poly`.
    poly_geobox = gbox.enclosing(poly)

    # Calculate ROI slices into `xx` for intersection between both geoboxes.
    roi = gbox.overlap_roi(poly_geobox)

    # Verify that `poly` overlaps with `xx` by checking if the returned
    # ROI is empty
    if roi_is_empty(roi):
        raise ValueError(
            "The supplied `poly` must overlap spatially with the extent of `xx`."
        )

    # Crop spatial dims of `xx` using ROI
    xx_cropped = xx.isel({sdims[0]: roi[0], sdims[1]: roi[1]})

    # Optionally mask data outside rasterized `poly`
    if apply_mask:
        xx_cropped = mask(xx_cropped, poly, all_touched=all_touched)

    return xx_cropped


def xr_coords(
    gbox: SomeGeoBox,
    crs_coord_name: Optional[str] = _DEFAULT_CRS_COORD_NAME,
    always_yx: bool = False,
    dims: Optional[Tuple[str, str]] = None,
) -> Dict[Hashable, xarray.DataArray]:
    """
    Dictionary of Coordinates in xarray format.

    :param gbox:
      :py:class:`~odc.geo.geobox.GeoBox` or :py:class:`~odc.geo.gcp.GCPGeoBox`

    :param crs_coord_name:
       Use custom name for CRS coordinate, default is "spatial_ref". Set to
       ``None`` to not generate CRS coordinate at all.

    :param always_yx:
       If True, always use names ``y,x`` for spatial coordinates even for
       geographic geoboxes.

    :param dims:
       Use custom names for spatial dimensions, default is to use ``y,x`` or
       ``latitude, longitude`` based on projection used. Dimensions are supplied
       in "array" order, i.e. ``('y', 'x')``.

    :returns:
       Dictionary ``name:str -> xr.DataArray``. Where names are either as
       supplied by ``dims=`` or otherwise ``y,x`` for projected or
       ``latitude, longitude`` for geographic.

    """
    if dims is None:
        if always_yx:
            dims = ("y", "x")
        else:
            dims = gbox.dimensions

    attrs = {}
    crs = gbox.crs
    if crs is not None:
        attrs["crs"] = str(crs)

    gcps = None
    transform: Optional[Affine] = None

    if isinstance(gbox, GCPGeoBox):
        coords: Dict[Hashable, xarray.DataArray] = {
            name: _mk_pixel_coord(name, sz) for name, sz in zip(dims, gbox.shape)
        }
        gcps = gbox.gcps()
    else:
        transform = gbox.transform
        if gbox.axis_aligned:
            coords = {
                name: _coord_to_xr(name, coord, **attrs)
                for name, coord in zip(dims, gbox.coordinates.values())
            }
        else:
            coords = {
                name: _mk_pixel_coord(name, sz) for name, sz in zip(dims, gbox.shape)
            }

    if crs_coord_name is not None and crs is not None:
        coords[crs_coord_name] = _mk_crs_coord(
            crs, crs_coord_name, gcps=gcps, transform=transform
        )

    return coords


def _mk_pixel_coord(
    name: str,
    sz: int,
) -> xarray.DataArray:
    data = numpy.arange(0.5, sz, dtype="float32")
    xx = xarray.DataArray(
        data, coords={name: data}, dims=(name,), attrs={"units": "pixel"}
    )
    return xx


def _is_spatial_ref(coord) -> bool:
    return coord.ndim == 0 and (
        "spatial_ref" in coord.attrs or "crs_wkt" in coord.attrs
    )


def _locate_crs_coords(xx: XarrayObject) -> List[xarray.DataArray]:
    grid_mapping = xx.encoding.get("grid_mapping", None)
    if grid_mapping is None:
        grid_mapping = xx.attrs.get("grid_mapping")

    if grid_mapping is not None:
        # Specific mapping is defined via NetCDF/CF convention
        coord = xx.coords.get(grid_mapping, None)
        if coord is None:
            warnings.warn(
                f"grid_mapping={grid_mapping} is not pointing to valid coordinate"
            )
            return []
        return [coord]

    # Find all dimensionless coordinates with `spatial_ref|crs_wkt` attribute present
    return [coord for coord in xx.coords.values() if _is_spatial_ref(coord)]


def _extract_crs(crs_coord: xarray.DataArray) -> Optional[CRS]:
    _wkt = crs_coord.attrs.get("spatial_ref", None)  # GDAL convention?
    if _wkt is None:
        _wkt = crs_coord.attrs.get("crs_wkt", None)  # CF convention
    if _wkt is None:
        return None
    try:
        return CRS(_wkt)
    except CRSError:
        return None


def _extract_gcps(crs_coord: xarray.DataArray) -> Optional[GCPMapping]:
    gcps = crs_coord.attrs.get("gcps", None)
    if gcps is None:
        return None
    crs = _extract_crs(crs_coord)
    try:
        if isinstance(gcps, str):
            gcps = json.loads(gcps)
        wld = Geometry(gcps, crs=crs)
        pix = [
            xy_(f["properties"]["col"], f["properties"]["row"])
            for f in gcps["features"]
        ]
        return GCPMapping(pix, wld)
    except (IndexError, KeyError, ValueError, json.JSONDecodeError):
        return None


def _extract_geo_transform(crs_coord: xarray.DataArray) -> Optional[Affine]:
    geo_transform_parts = crs_coord.attrs.get("GeoTransform", "").split(" ")
    if len(geo_transform_parts) != 6:
        return None
    try:
        c, a, b, f, d, e = map(float, geo_transform_parts)
    except ValueError:
        return None

    return Affine.from_gdal(c, a, b, f, d, e)


def _render_geo_transform(transform: Affine, precision: int = 24) -> str:
    return " ".join(
        map(lambda x: f"{x:.{precision}f}".rstrip("0").rstrip("."), transform.to_gdal())
    )


def _extract_transform(
    src: XarrayObject,
    sdims: Tuple[str, str],
    crs_coord: Optional[xarray.DataArray],
    gcp: bool,
) -> Optional[Affine]:
    if any(dim not in src.coords for dim in sdims):
        # special case of no spatial dims at all
        # happens for GCP/rotated sources loaded by rioxarray
        if gcp or crs_coord is None:
            return None
        return _extract_geo_transform(crs_coord)

    _yy, _xx = (src[dim] for dim in sdims)
    original_transform: Affine | None = None
    if crs_coord is not None:
        original_transform = _extract_geo_transform(crs_coord)

    # First try to compute from 1-D X/Y coords
    try:
        transform = affine_from_axis(_xx.values, _yy.values)
    except ValueError:
        # This can fail when any dimension is shorter than 2 elements
        # Figure out fallback resolution if possible and try again
        if crs_coord is None or original_transform is None:
            return None
        try:
            transform = affine_from_axis(
                _xx.values,
                _yy.values,
                resolution_from_affine(original_transform),
            )
        except ValueError:
            return None

    if original_transform is not None:
        if not is_affine_st(original_transform):
            # non-axis aligned geobox detected
            # adjust transform
            #  world <- pix' <- pix
            transform = original_transform * transform

        if any(map(math.isnan, transform)):
            transform = original_transform

        if approx_equal_affine(transform, original_transform):
            transform = original_transform

    return transform


def _locate_geo_info(src: XarrayObject) -> GeoState:
    # pylint: disable=too-many-locals
    if len(src.dims) < 2:
        return GeoState()

    sdims = spatial_dims(src, relaxed=True)
    if sdims is None:
        return GeoState()

    crs_coord: Optional[xarray.DataArray] = None
    crs: Optional[CRS] = None
    geobox: Optional[SomeGeoBox] = None
    gcp: Optional[GCPMapping] = None

    ny, nx = (src.coords[dim].shape[0] for dim in sdims)

    _crs_coords = _locate_crs_coords(src)
    num_candidates = len(_crs_coords)
    if num_candidates > 0:
        if num_candidates > 1:
            warnings.warn("Multiple CRS coordinates are present")
        crs_coord = _crs_coords[0]
        crs = _extract_crs(crs_coord)
        gcp = _extract_gcps(crs_coord)
    else:
        # try looking in attributes
        crs = _get_crs_from_attrs(src, sdims)

    transform = _extract_transform(src, sdims, crs_coord, gcp is not None)

    if gcp is not None:
        geobox = GCPGeoBox((ny, nx), gcp, transform)
    elif transform is not None:
        geobox = GeoBox((ny, nx), transform, crs)

    return GeoState(
        spatial_dims=sdims,
        crs_coord=crs_coord,
        transform=transform,
        crs=crs,
        geobox=geobox,
        gcp=gcp,
    )


def _wrap_op(method: F) -> F:
    @functools.wraps(method, assigned=("__doc__",))
    def wrapped(*args, **kw):
        # pylint: disable=protected-access
        _self, *rest = args
        return method(_self._xx, *rest, **kw)

    return wrapped  # type: ignore


def xr_reproject(
    src: XrT,
    how: Union[SomeCRS, GeoBox],
    *,
    resampling: Union[str, int] = "nearest",
    dst_nodata: SomeNodata = "auto",
    dtype=None,
    resolution: Union[SomeResolution, Literal["auto", "fit", "same"]] = "auto",
    shape: Union[SomeShape, int, None] = None,
    tight: bool = False,
    anchor: GeoboxAnchor = "default",
    tol: float = 0.01,
    round_resolution: Union[None, bool, Callable[[float, str], float]] = None,
    **kw,
) -> XrT:
    """
    Reproject raster to different projection/resolution.

    :param src:
      :py:class:`~xarray.Dataset` or :py:class:`~xarray.DataArray` to reproject.

    :param how:
      How to reproject the raster. Can be a GeoBox or a CRS (e.g. CRS object or
      an "ESPG:XXXX" string/integer). If a CRS is provided, the output pixel
      grid can be customised further via ``resolution``, ``shape``, ``tight``,
      ``anchor``, ``tol``, ``round_resolution``.

    :param resampling:
      Resampling method to use when reprojecting the raster. Defaults to
      "nearest", also supports "average", "bilinear", "cubic", "cubic_spline",
      "lanczos", "mode", "gauss", "max", "min", "med", "q1", "q3".

    :param dst_nodata:
      Set a custom nodata value for the output resampled raster.

    :param resolution:

       * "same" use exactly the same resolution as src
       * "fit" use center pixel to determine scale change between the two
       * | "auto" is to use the same resolution on the output if CRS units are
         | the same between the source and destination and otherwise use "fit"
       * Ignored if ``shape=`` is supplied
       * Else resolution in the units of the output crs

    :param shape:
      Span that many pixels, if it's a single number then span that many pixels
      along the longest dimension, other dimension will be computed to maintain
      roughly square pixels. Takes precedence over ``resolution=`` parameter.

    :param tight:
      By default output pixel grid is adjusted to align pixel edges to X/Y axis,
      suppling ``tight=True`` produces unaligned geobox on the output.

    :param anchor:
      Control pixel snapping, default is to snap pixel edge to ``X=0,Y=0``.
      Ignored when ``tight=True`` is supplied.

    :param tol:
       Fraction of the output pixel that can be ignored, defaults to 1/100.
       Bounding box of the output geobox is allowed to be smaller by that amount
       than transformed footprint of the original.

    :param round_resolution:
      ``round_resolution(res: float, units: str) -> float``

    This method uses :py:mod:`rasterio`.

    .. seealso:: :py:meth:`odc.geo.overlap.compute_output_geobox`

    """
    kw = {
        "shape": shape,
        "resolution": resolution,
        "tight": tight,
        "anchor": anchor,
        "tol": tol,
        "round_resolution": round_resolution,
        **kw,
    }
    if isinstance(src, xarray.DataArray):
        return _xr_reproject_da(
            src, how, resampling=resampling, dst_nodata=dst_nodata, dtype=dtype, **kw
        )
    return _xr_reproject_ds(
        src, how, resampling=resampling, dst_nodata=dst_nodata, dtype=dtype, **kw
    )


def _extract_output_geobox_params(kw):
    # NOTE: modifies input, removes keys
    out = {}
    for k in ("tight", "anchor", "resolution", "shape", "tol", "round_resolution"):
        if k in kw:
            out[k] = kw.pop(k)
    return out


def _xr_reproject_ds(
    src: Any,
    how: Union[SomeCRS, GeoBox],
    *,
    resampling: Union[str, int] = "nearest",
    dst_nodata: SomeNodata = "auto",
    dtype=None,
    **kw,
) -> xarray.Dataset:
    assert isinstance(src, xarray.Dataset)

    if have.rasterio is False:  # pragma: nocover
        raise RuntimeError("Please install `rasterio` to use this method")

    assert isinstance(src.odc, ODCExtensionDs)
    if src.odc.geobox is None:
        raise ValueError("Can not reproject non-georegistered array.")

    kw_gbox = _extract_output_geobox_params(kw)

    if isinstance(how, GeoBox):
        dst_geobox = how
    else:
        dst_geobox = src.odc.output_geobox(how, **kw_gbox)

    def _maybe_reproject(dv: xarray.DataArray):
        if dv.odc.geobox is None:
            # pass-through data variables without a geobox
            strip_coords = [str(c.name) for c in _locate_crs_coords(dv)]
            if len(strip_coords) > 0:
                dv = dv.drop_vars(strip_coords)
            return dv
        return _xr_reproject_da(
            dv,
            how=dst_geobox,
            resampling=resampling,
            dst_nodata=dst_nodata,
            dtype=dtype,
            **kw,
        )

    return src.map(_maybe_reproject)


def _xr_reproject_da(
    src: Any,
    how: Union[SomeCRS, GeoBox],
    *,
    resampling: Union[str, int] = "nearest",
    dst_nodata: SomeNodata = "auto",
    dtype=None,
    **kw,
) -> xarray.DataArray:
    # pylint: disable=too-many-locals
    assert isinstance(src, xarray.DataArray)

    if have.rasterio is False:  # pragma: nocover
        raise RuntimeError("Please install `rasterio` to use this method")

    assert isinstance(src.odc, ODCExtensionDa)  # for mypy sake
    src_gbox = src.odc.geobox

    if src_gbox is None or src_gbox.crs is None:
        raise ValueError("Can not reproject non-georegistered array.")

    kw_gbox = _extract_output_geobox_params(kw)

    if isinstance(how, GeoBox):
        dst_geobox = how
    else:
        dst_geobox = src.odc.output_geobox(how, **kw_gbox)

    if dtype is None:
        dtype = src.dtype

    # compute destination shape by replacing spatial dimensions shape
    ydim = src.odc.ydim
    assert ydim + 1 == src.odc.xdim
    dst_shape = (*src.shape[:ydim], *dst_geobox.shape, *src.shape[ydim + 2 :])

    src_nodata = resolve_nodata(kw.pop("src_nodata", "auto"), src.dtype, src.odc.nodata)
    dst_nodata = resolve_nodata(dst_nodata, dtype, src_nodata)

    fill_value = resolve_fill_value(dst_nodata, src_nodata, dtype)

    if is_dask_collection(src):
        from ._dask import dask_rio_reproject

        dst: Any = dask_rio_reproject(
            src.data,
            src_gbox,
            dst_geobox,
            resampling=resampling,
            src_nodata=src_nodata,
            dst_nodata=fill_value,
            ydim=ydim,
            dtype=dtype,
            **kw,
        )
    else:
        dst = numpy.full(dst_shape, fill_value, dtype=dtype)

        dst = rio_reproject(
            src.values,
            dst,
            src_gbox,
            dst_geobox,
            resampling=resampling,
            src_nodata=src_nodata,
            dst_nodata=fill_value,
            ydim=ydim,
            dtype=dtype,
            **kw,
        )

    attrs = {k: v for k, v in src.attrs.items() if k not in REPROJECT_SKIP_ATTRS}
    if not is_nodata_empty(dst_nodata):
        assert dst_nodata is not None
        attrs.update({k: maybe_int(float(dst_nodata), 1e-6) for k in NODATA_ATTRIBUTES})

    # new set of coords (replace x,y dims)
    # discard all coords that reference spatial dimensions
    sdims = src.odc.spatial_dims
    assert sdims is not None
    sdims = set(sdims)

    def should_keep(coord):
        if _is_spatial_ref(coord):
            return False
        return sdims.isdisjoint(coord.dims)

    coords = {k: coord for k, coord in src.coords.items() if should_keep(coord)}
    coords.update(xr_coords(dst_geobox))

    dims = (*src.dims[:ydim], *dst_geobox.dimensions, *src.dims[ydim + 2 :])

    out = xarray.DataArray(dst, coords=coords, dims=dims, attrs=attrs)
    out.encoding["grid_mapping"] = _DEFAULT_CRS_COORD_NAME
    return out


class ODCExtension:
    """
    ODC extension base class.

    Common accessors for both Array/Dataset.
    """

    def __init__(self, state: GeoState):
        self._state = state

    @property
    def spatial_dims(self) -> Optional[Tuple[str, str]]:
        """Return names of spatial dimensions, or ``None``."""
        return self._state.spatial_dims

    @property
    def transform(self) -> Optional[Affine]:
        return self._state.transform

    affine = transform

    @property
    def crs(self) -> Optional[CRS]:
        """Query :py:class:`~odc.geo.crs.CRS`."""
        return self._state.crs

    @property
    def geobox(self) -> Optional[SomeGeoBox]:
        """Query :py:class:`~odc.geo.geobox.GeoBox` or :py:class:`~odc.geo.gcp.GCPGeoBox`."""
        return self._state.geobox

    @property
    def aspect(self) -> float:
        gbox = self._state.geobox
        if gbox is None:
            return 1
        return gbox.aspect

    def output_geobox(self, crs: SomeCRS, **kw) -> GeoBox:
        """
        Compute geobox of this data in other projection.

        .. seealso:: :py:meth:`odc.geo.overlap.compute_output_geobox`
        """
        gbox = self.geobox
        if gbox is None:
            raise ValueError("Not geo registered")

        return compute_output_geobox(gbox, crs, **kw)

    def map_bounds(self) -> Tuple[Tuple[float, float], Tuple[float, float]]:
        """See :py:meth:`odc.geo.geobox.GeoBox.map_bounds`."""
        gbox = self.geobox
        if gbox is None:
            raise ValueError("Not geo registered")

        return gbox.map_bounds()

    @property
    def crs_coord(self) -> xarray.DataArray | None:
        """Return CRS coordinate DataArray."""
        return self._state.crs_coord

    @property
    def grid_mapping(self) -> str | None:
        """Return name of the grid mapping coordinate."""
        if c := self.crs_coord:
            return str(c.name)
        return None

    mask = _wrap_op(mask)
    crop = _wrap_op(crop)

    if have.rasterio:
        explore = _wrap_op(explore)
        reproject = _wrap_op(xr_reproject)


@xarray.register_dataarray_accessor("odc")
class ODCExtensionDa(ODCExtension):
    """
    ODC extension for :py:class:`xarray.DataArray`.
    """

    def __init__(self, xx: xarray.DataArray):
        ODCExtension.__init__(self, _locate_geo_info(xx))
        self._xx = xx

    @property
    def uncached(self) -> "ODCExtensionDa":
        return ODCExtensionDa(self._xx)

    def reload(self) -> xarray.DataArray:
        """Reload geospatial state info in-place."""
        self._state = _locate_geo_info(self._xx)
        return self._xx

    @property
    def ydim(self) -> int:
        """Index of the Y dimension."""
        if (sdims := self.spatial_dims) is not None:
            return self._xx.dims.index(sdims[0])
        raise ValueError("Can't locate spatial dimensions")

    @property
    def xdim(self) -> int:
        """Index of the X dimension."""
        if (sdims := self.spatial_dims) is not None:
            return self._xx.dims.index(sdims[1])
        raise ValueError("Can't locate spatial dimensions")

    def assign_crs(
        self, crs: SomeCRS, crs_coord_name: str = _DEFAULT_CRS_COORD_NAME
    ) -> xarray.DataArray:
        """See :py:meth:`odc.geo.xr.assign_crs`."""
        return assign_crs(self._xx, crs=crs, crs_coord_name=crs_coord_name)

    @property
    def nodata(self) -> Nodata:
        """Extract ``nodata/_FillValue`` attribute if set."""
        attrs = self._xx.attrs
        encoding = self._xx.encoding

        for k in ["nodata", "_FillValue"]:
            nodata = attrs.get(k, _NoValue)
            if nodata is _NoValue:
                nodata = encoding.get(k, _NoValue)

            if nodata is _NoValue:
                continue

            if nodata is None:
                return None

            return float(nodata)

        return None

    @nodata.setter
    def nodata(self, value: Nodata):
        nodata = resolve_nodata(value, self._xx.dtype)

        if nodata is None:
            for k in ["nodata", "_FillValue"]:
                self._xx.attrs.pop(k, None)
                self._xx.encoding.pop(k, None)
            return

        self._xx.attrs["nodata"] = nodata
        self._xx.encoding["_FillValue"] = nodata

    colorize = _wrap_op(colorize)

    if have.rasterio:
        write_cog = _wrap_op(write_cog)
        to_cog = _wrap_op(to_cog)
        compress = _wrap_op(compress)
        add_to = _wrap_op(add_to)


@xarray.register_dataset_accessor("odc")
class ODCExtensionDs(ODCExtension):
    """
    ODC extension for :py:class:`xarray.Dataset`.
    """

    def __init__(self, ds: xarray.Dataset):
        ODCExtension.__init__(self, _locate_geo_info(ds))
        self._xx = ds

    def reload(self) -> xarray.Dataset:
        """Reload geospatial state info in-place."""
        self._state = _locate_geo_info(self._xx)
        return self._xx

    @property
    def uncached(self) -> "ODCExtensionDs":
        return ODCExtensionDs(self._xx)

    def assign_crs(
        self, crs: SomeCRS, crs_coord_name: str = _DEFAULT_CRS_COORD_NAME
    ) -> xarray.Dataset:
        return assign_crs(self._xx, crs=crs, crs_coord_name=crs_coord_name)

    def to_rgba(
        self,
        bands: Optional[Tuple[str, str, str]] = None,
        *,
        vmin: Optional[float] = None,
        vmax: Optional[float] = None,
    ) -> xarray.DataArray:
        return to_rgba(self._xx, bands=bands, vmin=vmin, vmax=vmax)


ODCExtensionDs.to_rgba.__doc__ = to_rgba.__doc__


def _xarray_geobox(xx: XarrayObject) -> Optional[GeoBox]:
    if isinstance(xx, xarray.DataArray):
        return xx.odc.geobox
    for dv in xx.data_vars.values():
        geobox = dv.odc.geobox
        if geobox is not None:
            return geobox
    return None


def register_geobox():
    """
    Backwards compatiblity layer for datacube ``.geobox`` property.
    """
    xarray.Dataset.geobox = property(_xarray_geobox)  # type: ignore
    xarray.DataArray.geobox = property(_xarray_geobox)  # type: ignore


def wrap_xr(
    im: Any,
    gbox: SomeGeoBox,
    *,
    time=None,
    nodata: SomeNodata = "auto",
    crs_coord_name: Optional[str] = _DEFAULT_CRS_COORD_NAME,
    always_yx: bool = False,
    dims: Optional[Tuple[str, ...]] = None,
    axis: Optional[int] = None,
    **attrs,
) -> xarray.DataArray:
    """
    Wrap xarray around numpy array with CRS and x,y coords.

    :param im: numpy array to wrap, last two axes are Y,X
    :param gbox: Geobox, must same shape as last two axis of ``im``
    :param time: optional time axis value(s), defaults to None
    :param nodata: optional `nodata` value, defaults to None
    :param crs_coord_name: allows to change name of the crs coordinate variable
    :param always_yx: If True, always use names ``y,x`` for spatial coordinates
    :param dims: Custom names for spatial dimensions
    :param axis: Which axis of the input array corresponds to Y,X
    :param attrs: Any other attributes to set on the result
    :return: xarray DataArray
    """
    # pylint: disable=too-many-locals,too-many-arguments
    assert dims is None or len(dims) == im.ndim

    if axis is None:
        axis = 1 if time is not None else 0
    elif axis < 0:  # handle numpy style negative axis
        axis = int(im.ndim) + axis

    if im.ndim == 2 and axis == 1:
        im = im[numpy.newaxis, ...]

    assert axis >= 0
    assert im.ndim - axis - 2 >= 0
    assert im.shape[axis : axis + 2] == gbox.shape

    def _prefix_dims(n):
        if n == 0:
            return ()
        if n == 1:
            return ("time",)
        return ("time", *[f"dim_{i}" for i in range(n - 1)])

    def _postfix_dims(n):
        if n == 0:
            return ()
        if n == 1:
            return ("band",)
        return (f"b_{i}" for i in range(n))

    sdims: Optional[Tuple[str, str]] = None
    if dims is None:
        sdims = ("y", "x") if always_yx else gbox.dimensions
        dims = (*_prefix_dims(axis), *sdims, *_postfix_dims(im.ndim - axis - 2))
    else:
        sdims = dims[axis], dims[axis + 1]

    prefix_dims = dims[:axis]
    postfix_dims = dims[axis + 2 :]

    coords = xr_coords(
        gbox,
        crs_coord_name=crs_coord_name,
        always_yx=always_yx,
        dims=sdims,
    )

    if time is not None:
        if not isinstance(time, xarray.DataArray):
            if len(prefix_dims) > 0 and isinstance(time, (str, datetime)):
                time = [time]

            time = xarray.DataArray(time, dims=prefix_dims[:1]).astype("datetime64[ns]")

        coords["time"] = time

    if postfix_dims:
        for a, dim in enumerate(postfix_dims):
            nb = im.shape[axis + 2 + a]
            coords[dim] = xarray.DataArray(
                [f"b{i}" for i in range(nb)], dims=(dim,), name=dim
            )

    _nodata = resolve_nodata(nodata, im.dtype)
    if not is_nodata_empty(_nodata) or nodata != "auto":
        attrs = {"nodata": _nodata, **attrs}

    out = xarray.DataArray(im, coords=coords, dims=dims, attrs=attrs)
    if crs_coord_name is not None:
        out.encoding["grid_mapping"] = crs_coord_name
    return out


def xr_zeros(
    geobox: SomeGeoBox,
    dtype="float64",
    *,
    chunks: Optional[Union[Tuple[int, int], Tuple[int, int, int]]] = None,
    time=None,
    crs_coord_name: Optional[str] = _DEFAULT_CRS_COORD_NAME,
    **kw,
) -> xarray.DataArray:
    """
    Construct geo-registered xarray from a :py:class:`~odc.geo.geobox.GeoBox`.

    :param gbox: Desired footprint and resolution
    :param dtype: Pixel data type
    :param chunks: Create a dask array instead of numpy array
    :param time: When set adds time dimension
    :param crs_coord_name: allows to change name of the crs coordinate variable

    :return: :py:class:`xarray.DataArray` filled with zeros (numpy or dask)

    .. seealso:: :py:meth:`odc.geo.xr.wrap_xr`

    """
    if time is not None:
        _shape: Tuple[int, ...] = (len(time), *geobox.shape.yx)
    else:
        _shape = geobox.shape.yx

    if chunks is not None:
        from dask import array as da  # pylint: disable=import-outside-toplevel

        return wrap_xr(
            da.zeros(_shape, dtype=dtype, chunks=chunks),
            geobox,
            crs_coord_name=crs_coord_name,
            time=time,
            **kw,
        )

    return wrap_xr(
        numpy.zeros(_shape, dtype=dtype),
        geobox,
        crs_coord_name=crs_coord_name,
        time=time,
        **kw,
    )


def rasterize(
    poly: Geometry,
    how: Union[float, int, Resolution, GeoBox],
    *,
    value_inside: bool = True,
    all_touched: bool = False,
) -> xarray.DataArray:
    """
    Generate raster from geometry.

    This method is a wrapper for :py:meth:`rasterio.features.make_mask`.

    :param poly:
       Geometry shape to rasterize.

    :param how:
        This could be either just resolution or a GeoBox that fully defines output
        raster extent/resolution/projection.

    :param all_touched:
        If ``True``, all pixels touched by geometries will be burned in.  If
        ``False``, only pixels whose center is within the polygon or that
        are selected by Bresenham's line algorithm will be burned in.

    :param value_inside:
        By default pixels inside a polygon will have value of ``True`` and ``False``
        outside, but this can be flipped.

    :return: geo-registered data array
    """
    # pylint: disable=import-outside-toplevel

    if have.rasterio is False:  # pragma: nocover
        raise RuntimeError("Please install `rasterio` to use this method")

    from rasterio.features import geometry_mask

    if isinstance(how, GeoBox):
        geobox = how
    else:
        geobox = GeoBox.from_geopolygon(poly, resolution=how)

    if poly.crs != geobox.crs and geobox.crs is not None:
        poly = poly.to_crs(geobox.crs)

    pix = geometry_mask(
        [poly.geom],
        geobox.shape,
        geobox.transform,
        all_touched=all_touched,
        invert=value_inside,
    )
    return wrap_xr(pix, geobox)



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2020 ODC Contributors
# SPDX-License-Identifier: Apache-2.0

from types import SimpleNamespace
from unittest.mock import Mock

import pytest
import rasterio.crs
from pytest import approx

from odc.geo import geom
from odc.geo.crs import (
    _CRS,
    CRS,
    CRSError,
    CRSMismatchError,
    crs_units_per_degree,
    norm_crs,
    norm_crs_or_error,
)
from odc.geo.geom import common_crs
from odc.geo.testutils import epsg3577, epsg3857, epsg4326
from odc.geo.types import Unset, xy_

# pylint: disable=missing-class-docstring,use-implicit-booleaness-not-comparison
# pylint: disable=comparison-with-itself


def test_common_crs():
    assert common_crs([]) is None
    assert (
        common_crs([geom.point(0, 0, epsg4326), geom.line([(0, 0), (1, 1)], epsg4326)])
        is epsg4326
    )

    with pytest.raises(CRSMismatchError):
        common_crs([geom.point(0, 0, epsg4326), geom.line([(0, 0), (1, 1)], epsg3857)])


class TestCRSEqualityComparisons:
    def test_comparison_edge_cases(self):
        a = epsg4326
        none_crs = None
        assert a == a
        assert a == str(a)
        assert (a == none_crs) is False
        assert (a == []) is False
        assert (a == TestCRSEqualityComparisons) is False

    def test_australian_albers_comparison(self):
        a = CRS(
            """PROJCS["GDA94_Australian_Albers",GEOGCS["GCS_GDA_1994",
                            DATUM["Geocentric_Datum_of_Australia_1994",SPHEROID["GRS_1980",6378137,298.257222101]],
                            PRIMEM["Greenwich",0],UNIT["Degree",0.017453292519943295]],
                            PROJECTION["Albers_Conic_Equal_Area"],
                            PARAMETER["standard_parallel_1",-18],
                            PARAMETER["standard_parallel_2",-36],
                            PARAMETER["latitude_of_center",0],
                            PARAMETER["longitude_of_center",132],
                            PARAMETER["false_easting",0],
                            PARAMETER["false_northing",0],
                            UNIT["Meter",1]]"""
        )
        b = epsg3577

        assert a == b

        assert a != epsg4326


def test_no_epsg():
    c = CRS("+proj=longlat +no_defs +ellps=GRS80")
    b = CRS(
        """GEOGCS["GRS 1980(IUGG, 1980)",DATUM["unknown",SPHEROID["GRS80",6378137,298.257222101]],
                        PRIMEM["Greenwich",0],UNIT["degree",0.0174532925199433]]"""
    )

    assert c.epsg is None
    assert b.epsg is None
    assert c.authority == ("", "")


def test_crs():
    custom_crs = CRS(
        """PROJCS["unnamed",
                           GEOGCS["Unknown datum based upon the custom spheroid",
                           DATUM["Not specified (based on custom spheroid)", SPHEROID["Custom spheroid",6371007.181,0]],
                           PRIMEM["Greenwich",0],UNIT["degree",0.0174532925199433]],
                           PROJECTION["Sinusoidal"],
                           PARAMETER["longitude_of_center",0],
                           PARAMETER["false_easting",0],
                           PARAMETER["false_northing",0],
                           UNIT["Meter",1]]"""
    )

    crs = epsg3577
    assert crs.geographic is False
    assert crs.projected is True
    assert crs.dimensions == ("y", "x")
    assert crs.epsg == 3577
    assert crs.units == ("metre", "metre")
    assert isinstance(repr(crs), str)

    crs = epsg4326
    assert crs.geographic is True
    assert crs.projected is False
    assert crs.dimensions == ("latitude", "longitude")
    assert crs.epsg == 4326
    assert epsg4326.semi_major_axis > 6_350_000
    assert epsg4326.semi_minor_axis > 6_350_000
    assert epsg4326.inverse_flattening == pytest.approx(298.257223563)

    crs2 = CRS(crs)
    assert crs2 == crs
    assert crs.proj is crs2.proj

    assert epsg4326.valid_region == geom.box(-180, -90, 180, 90, epsg4326)
    assert epsg3857.valid_region.crs == epsg4326
    xmin, _, xmax, _ = epsg3857.valid_region.boundingbox
    assert (xmin, xmax) == (-180, 180)
    assert custom_crs.valid_region is None

    assert epsg3577 == epsg3577
    assert epsg3577 == "EPSG:3577"
    assert (epsg3577 != epsg3577) is False
    assert (epsg3577 == epsg4326) is False
    assert (epsg3577 == "EPSG:4326") is False
    assert epsg3577 != epsg4326
    assert epsg3577 != "EPSG:4326"

    bad_crs = [
        "cupcakes",
        (
            'PROJCS["unnamed",'
            'GEOGCS["WGS 84", DATUM["WGS_1984", SPHEROID["WGS 84",6378137,298.257223563, AUTHORITY["EPSG","7030"]],'
            'AUTHORITY["EPSG","6326"]], PRIMEM["Greenwich",0, AUTHORITY["EPSG","8901"]],'
            'UNIT["degree",0.0174532925199433, AUTHORITY["EPSG","9122"]], AUTHORITY["EPSG","4326"]]]'
        ),
    ]

    for bad in bad_crs:
        with pytest.raises(CRSError):
            CRS(bad)

    with pytest.warns(DeprecationWarning):
        assert str(epsg3857) == epsg3857.crs_str


def test_crs_compat():
    crs = CRS("epsg:3577")
    assert crs.epsg == 3577
    assert crs.authority == ("EPSG", 3577)
    assert str(crs) == "EPSG:3577"
    crs2 = CRS(crs)
    assert crs.epsg == crs2.epsg

    assert crs == CRS(crs.proj)
    assert crs == CRS(crs)
    assert crs == CRS(f"epsg:{crs.to_epsg()}")
    assert crs == CRS(crs.to_wkt())

    assert str(CRS(crs.to_wkt())) == "EPSG:3577"

    crs_rio = rasterio.crs.CRS(init="epsg:3577")
    assert CRS(crs_rio).epsg == 3577

    assert (CRS(crs_rio) == crs_rio) is True
    assert CRS(_CRS(crs_rio)) == CRS(crs_rio)
    assert CRS(str(crs_rio)) == crs_rio
    assert CRS(crs_rio) == 3577

    with pytest.raises(CRSError):
        CRS(("random", "tuple"))

    crs = CRS("epsg:3857")
    with pytest.warns(UserWarning):
        crs_dict = crs.proj.to_dict()

    assert CRS(crs_dict) == crs

    wkt_crs = SimpleNamespace(to_wkt=Mock(return_value=crs.to_wkt()))
    assert getattr(wkt_crs, "to_epsg", None) is None
    assert CRS(wkt_crs) == crs

    assert CRS("ESRI:54019").authority == ("ESRI", 54019)


@pytest.mark.parametrize("epsg", [4326, 3577, 3587])
def test_crs_int(epsg):
    assert CRS(epsg) == CRS(f"EPSG:{epsg}")
    assert CRS(epsg).epsg == epsg

    crs1 = CRS(epsg)
    crs2 = CRS(str(crs1))
    crs3 = CRS(f"epsg:{epsg}")
    assert crs1.proj is crs2.proj
    assert crs1.proj is crs3.proj


def test_crs_hash():
    crs = CRS("epsg:3577")
    crs2 = CRS(crs)

    assert crs is not crs2
    assert len({crs, crs2}) == 1


def test_crs_units_per_degree():
    assert crs_units_per_degree("EPSG:3857", (0, 0)) == crs_units_per_degree(
        "EPSG:3857", 0, 0
    )
    assert crs_units_per_degree("EPSG:4326", (120, -10)) == approx(1.0, 1e-6)

    assert crs_units_per_degree("EPSG:3857", 0, 0) == approx(111319.49, 0.5)
    assert crs_units_per_degree("EPSG:3857", 20, 0) == approx(111319.49, 0.5)
    assert crs_units_per_degree("EPSG:3857", 30, 0) == approx(111319.49, 0.5)
    assert crs_units_per_degree("EPSG:3857", 180, 0) == approx(111319.49, 0.5)
    assert crs_units_per_degree("EPSG:3857", -180, 0) == approx(111319.49, 0.5)


def test_rio_crs__no_epsg():
    rio_crs = rasterio.crs.CRS.from_wkt(
        'PROJCS["unnamed",GEOGCS["Unknown datum based upon the custom spheroid",'
        'DATUM["Not specified (based on custom spheroid)",'
        'SPHEROID["Custom spheroid",6371007.181,0]],'
        'PRIMEM["Greenwich",0],UNIT["degree",0.0174532925199433]],'
        'PROJECTION["Sinusoidal"],PARAMETER["longitude_of_center",0],'
        'PARAMETER["false_easting",0],PARAMETER["false_northing",0],'
        'UNIT["Meter",1],AXIS["Easting",EAST],AXIS["Northing",NORTH]]'
    )
    assert CRS(rio_crs).epsg is None


def test_norm_crs():
    assert norm_crs(None) is None
    assert norm_crs(Unset()) is None

    for v in [None, Unset()]:
        with pytest.raises(ValueError):
            _ = norm_crs_or_error(v)


@pytest.mark.parametrize(
    "x, expected_epsg",
    [
        (-180, 32601),
        ((-180, -30), 32701),
        (xy_(0.33, 10), 32631),
        (xy_(180, 10), 32660),
        (geom.point(6.89, 50.3, "epsg:4326"), 32632),
        (geom.point(6.89, -50.3, "epsg:4326"), 32732),
        (geom.point(6.89, 50.3, None), 32632),
        (geom.point(6.89, 50.3, "epsg:4326").to_crs("epsg:3857"), 32632),
        (geom.point(6.89, 50.3, "epsg:4326").to_crs("epsg:3857").buffer(1000), 32632),
        (geom.BoundingBox(0, 10, 7, 20), 32631),
        (geom.BoundingBox(0, 10, 7, 20, "epsg:4326"), 32631),
        (geom.BoundingBox(5, 10, 8, 20), 32632),
    ],
)
def test_crs_utm(x, expected_epsg):
    if isinstance(x, tuple):
        crs = CRS.utm(*x)
    else:
        crs = CRS.utm(x)

    assert crs.epsg is not None
    assert crs.epsg == expected_epsg
    if isinstance(x, tuple):
        return

    assert norm_crs_or_error("utm", x).epsg == expected_epsg
    assert norm_crs_or_error("UTM", x).epsg == expected_epsg
    assert norm_crs("UTM", x).epsg == expected_epsg

    if crs.proj.utm_zone.endswith("S"):
        assert norm_crs("utm-s", x) == norm_crs("utm", x)
        assert norm_crs("utm-n", x).epsg == expected_epsg - 100
    elif crs.proj.utm_zone.endswith("N"):
        assert norm_crs("utM-n", x) == norm_crs("utm", x)
        assert norm_crs("uTm-s", x).epsg == expected_epsg + 100

    with pytest.raises(ValueError):
        _ = CRS.utm(x, datum_name="no such datum")


def test_crs_units_issue_120():
    compound_crs_wkt = """COMPOUNDCRS["GDA94 / MGA zone 50 + Instantaneous Water Level Height",
    PROJCRS["GDA94 / MGA zone 50",
        BASEGEOGCRS["GDA94",
            DATUM["Geocentric Datum of Australia 1994",
                ELLIPSOID["GRS 1980",6378137,298.257222101,
                    LENGTHUNIT["metre",1]]],
            PRIMEM["Greenwich",0,
                ANGLEUNIT["degree",0.0174532925199433]],
            ID["EPSG",4283]],
        CONVERSION["UTM zone 50S",
            METHOD["Transverse Mercator",
                ID["EPSG",9807]],
            PARAMETER["Latitude of natural origin",0,
                ANGLEUNIT["degree",0.0174532925199433],
                ID["EPSG",8801]],
            PARAMETER["Longitude of natural origin",117,
                ANGLEUNIT["degree",0.0174532925199433],
                ID["EPSG",8802]],
            PARAMETER["Scale factor at natural origin",0.9996,
                SCALEUNIT["unity",1],
                ID["EPSG",8805]],
            PARAMETER["False easting",500000,
                LENGTHUNIT["metre",1],
                ID["EPSG",8806]],
            PARAMETER["False northing",10000000,
                LENGTHUNIT["metre",1],
                ID["EPSG",8807]]],
        CS[Cartesian,2],
            AXIS["easting",east,
                ORDER[1],
                LENGTHUNIT["metre",1]],
            AXIS["northing",north,
                ORDER[2],
                LENGTHUNIT["metre",1]],
        ID["EPSG",28350]],
    VERTCRS["Instantaneous Water Level Height",
        VDATUM["Instantaneous Water Level"],
        CS[vertical,1],
            AXIS["up",up,
                LENGTHUNIT["metre",1]],
        ID["EPSG",5829]]]"""
    assert CRS(compound_crs_wkt).units == ("metre", "metre")



import numpy as np
import pytest
import xarray as xr

from odc.geo._interop import have
from odc.geo.xr import ODCExtensionDa

# pylint: disable=protected-access,import-outside-toplevel

cmap = np.asarray(
    [
        [153, 153, 102, 255],
        [51, 102, 153, 255],
    ],
    dtype="uint8",
)


def test_add_to(ocean_raster: xr.DataArray):
    # Check map safe range first
    xx = ocean_raster.copy()[20:-20]
    assert isinstance(xx.odc, ODCExtensionDa)

    url, bounds = xx.odc.add_to(None, cmap=cmap, max_size=32, zlevel=1)
    assert isinstance(url, str)
    assert url.startswith("data:")
    assert "image/png" in url
    assert isinstance(bounds, tuple)
    assert isinstance(bounds[0], tuple)
    assert isinstance(bounds[0], tuple)

    url, bounds = xx.odc.add_to(None, cmap=cmap, max_size=32, fmt="jpeg", quality=50)
    assert isinstance(url, str)
    assert url.startswith("data:")
    assert "image/jpeg" in url
    assert isinstance(bounds, tuple)
    assert isinstance(bounds[0], tuple)
    assert isinstance(bounds[0], tuple)

    for not_a_map in [xx.data, {}, []]:
        with pytest.raises(ValueError):
            _ = xx.odc.add_to(not_a_map)


@pytest.mark.skipif(have.folium is False, reason="No folium installed")
def test_add_to_folium(ocean_raster: xr.DataArray):
    import folium

    xx = ocean_raster.copy()[20:-20]
    assert isinstance(xx.odc, ODCExtensionDa)

    m = folium.Map()
    img_overlay = xx.odc.add_to(m, cmap=cmap, max_size=32, zlevel=1)
    assert img_overlay is not None


@pytest.mark.skipif(have.ipyleaflet is False, reason="No ipyleaflet installed")
def test_add_to_ipyleaflet(ocean_raster: xr.DataArray):
    import ipyleaflet

    xx = ocean_raster.copy()[20:-20]
    assert isinstance(xx.odc, ODCExtensionDa)

    m = ipyleaflet.Map()
    img_overlay = xx.odc.add_to(m, cmap=cmap, max_size=32, name="xx")
    assert img_overlay is not None


@pytest.mark.skipif(have.folium is False, reason="No folium installed")
def test_explore_xr(ocean_raster: xr.DataArray, ocean_raster_ds: xr.Dataset):
    import folium
    from folium.raster_layers import ImageOverlay

    # Test explore on dataset input and verify that output is a folium map
    # that contains an ImageOverlay layer
    m = ocean_raster_ds.odc.explore()
    assert isinstance(m, folium.folium.Map)
    assert any(isinstance(child, ImageOverlay) for child in m._children.values())

    # Verify that explore fails if bands cannot be guessed
    with pytest.raises(ValueError):
        ocean_raster_ds.rename({"red": "band1"}).odc.explore()

    # Verify that error can be avoided using `bands` param
    ocean_raster_ds.rename({"red": "band1"}).odc.explore(
        bands=("band1", "green", "blue")
    )

    # Test explore on data array input and verify that output is a folium
    # map that contains an ImageOverlay layer
    m = ocean_raster.odc.explore()
    assert isinstance(m, folium.folium.Map)
    assert any(isinstance(child, ImageOverlay) for child in m._children.values())

    # Verify that error is raised if a timeseries is passed
    with pytest.raises(ValueError):
        xr.concat([ocean_raster, ocean_raster], dim="time").odc.explore()



import base64
from typing import Any

from ._multipart import MultiPartUploadBase


class AzureLimits:
    """
    Common Azure writer settings.
    """

    @property
    def min_write_sz(self) -> int:
        # Azure minimum write size for blocks (default is 4 MiB)
        return 4 * (1 << 20)

    @property
    def max_write_sz(self) -> int:
        # Azure maximum write size for blocks (default is 100 MiB)
        return 100 * (1 << 20)

    @property
    def min_part(self) -> int:
        return 1

    @property
    def max_part(self) -> int:
        # Azure supports up to 50,000 blocks per blob
        return 50_000


class AzMultiPartUpload(AzureLimits, MultiPartUploadBase):
    """
    Azure Blob Storage multipart upload.
    """

    # pylint: disable=too-many-instance-attributes

    def __init__(
        self, account_url: str, container: str, blob: str, credential: Any = None
    ):
        """
        Initialise Azure multipart upload.

        :param account_url: URL of the Azure storage account.
        :param container: Name of the container.
        :param blob: Name of the blob.
        :param credential: Authentication credentials (e.g., SAS token or key).
        """
        self.account_url = account_url
        self.container = container
        self.blob = blob
        self.credential = credential

        # Initialise Azure Blob service client
        # pylint: disable=import-outside-toplevel,import-error
        from azure.storage.blob import BlobServiceClient

        self.blob_service_client = BlobServiceClient(
            account_url=account_url, credential=credential
        )
        self.container_client = self.blob_service_client.get_container_client(container)
        self.blob_client = self.container_client.get_blob_client(blob)

        self.block_ids: list[str] = []

    def initiate(self, **kwargs) -> str:
        """
        Initialise the upload. No-op for Azure.
        """
        return "azure-block-upload"

    def write_part(self, part: int, data: bytes) -> dict[str, Any]:
        """
        Stage a block in Azure.

        :param part: Part number (unique).
        :param data: Data for this part.
        :return: A dictionary containing part information.
        """
        block_id = base64.b64encode(f"block-{part}".encode()).decode()
        self.blob_client.stage_block(block_id=block_id, data=data)
        self.block_ids.append(block_id)
        return {"PartNumber": part, "BlockId": block_id}

    def finalise(self, parts: list[dict[str, Any]]) -> str:
        """
        Commit the block list to finalise the upload.

        :param parts: List of uploaded parts metadata.
        :return: The ETag of the finalised blob.
        """
        # pylint: disable=import-outside-toplevel,import-error
        from azure.storage.blob import BlobBlock

        block_list = [BlobBlock(block_id=part["BlockId"]) for part in parts]
        self.blob_client.commit_block_list(block_list)
        return self.blob_client.get_blob_properties().etag

    def cancel(self, other: str = ""):
        """
        Cancel the upload by clearing the block list.
        """
        assert other == ""
        self.block_ids.clear()

    @property
    def url(self) -> str:
        """
        Get the Azure blob URL.

        :return: The full URL of the blob.
        """
        return self.blob_client.url

    @property
    def started(self) -> bool:
        """
        Check if any blocks have been staged.

        :return: True if blocks have been staged, False otherwise.
        """
        return bool(self.block_ids)

    def writer(self, kw: dict[str, Any], *, client: Any = None):
        """
        Return a stateless writer compatible with Dask.
        """
        return DelayedAzureWriter(self, kw)

    def dask_name_prefix(self) -> str:
        """Return the Dask name prefix for Azure."""
        return "azure-finalise"


class DelayedAzureWriter(AzureLimits):
    """
    Dask-compatible writer for Azure Blob Storage multipart uploads.
    """

    def __init__(self, mpu: AzMultiPartUpload, kw: dict[str, Any]):
        """
        Initialise the Azure writer.

        :param mpu: AzMultiPartUpload instance.
        :param kw: Additional parameters for the writer.
        """
        self.mpu = mpu
        self.kw = kw  # Additional metadata like ContentType

    def __call__(self, part: int, data: bytes) -> dict[str, Any]:
        """
        Write a single part to Azure Blob Storage.

        :param part: Part number.
        :param data: Chunk data.
        :return: Metadata for the written part.
        """
        return self.mpu.write_part(part, data)

    def finalise(self, parts: list[dict[str, Any]]) -> str:
        """
        Finalise the upload by committing the block list.

        :param parts: List of uploaded parts metadata.
        :return: ETag of the finalised blob.
        """
        return self.mpu.finalise(parts)



# pylint: disable=wrong-import-position, protected-access, redefined-outer-name
from pathlib import Path

import numpy as np
import pytest

from odc.geo import geom
from odc.geo.gcp import GCPGeoBox, GCPMapping
from odc.geo.geobox import AnchorEnum, GeoBox
from odc.geo.xr import xr_zeros

rasterio = pytest.importorskip("rasterio")


@pytest.fixture()
def au_gcp_rio(data_dir: Path):
    with rasterio.open(data_dir / "au-gcp.tif") as src:
        yield src


@pytest.fixture()
def au_gcp_geobox(au_gcp_rio):
    yield GCPGeoBox.from_rio(au_gcp_rio)


def test_gcp_geobox_from_rio(au_gcp_rio):
    src = au_gcp_rio
    assert src.gcps != ([], None)
    gbox = GCPGeoBox.from_rio(src)
    assert gbox.crs == "epsg:4326"

    assert src.width == gbox.width
    assert src.height == gbox.height


def test_gcp_geobox_basics(au_gcp_geobox: GCPGeoBox):
    gbox = au_gcp_geobox

    assert gbox.linear is False
    assert gbox.axis_aligned is False

    assert gbox.pad(1)[1:-1, 1:-1] == gbox
    assert gbox.pad(1).width == gbox.width + 2
    assert gbox.pad(-1).height == gbox.height - 2

    assert (gbox.extent - gbox.pad(2).extent).is_empty
    assert gbox.geographic_extent.crs == "epsg:4326"

    assert gbox.boundary().shape[1] == 2

    assert gbox.approx.shape == gbox.shape
    assert isinstance(gbox.approx, GeoBox)

    assert gbox.resolution.xy == pytest.approx((0.308896, -0.269107), abs=1e-5)
    assert gbox.resolution == gbox._mapping.resolution

    g2 = gbox.pad_wh(10)
    assert g2.width % 10 == 0 and g2.width >= gbox.width
    assert g2.height % 10 == 0 and g2.height >= gbox.height
    assert g2[0, 0] == gbox[0, 0]
    assert str(gbox) != str(g2)
    assert g2.resolution == gbox.resolution

    assert gbox.center_pixel == gbox[gbox.height // 2, gbox.width // 2]

    assert str(gbox) == repr(gbox)
    assert gbox != "some other thing"

    assert gbox.wld2pix(*gbox.pix2wld(0, 0)) == pytest.approx((0, 0), abs=1)
    assert gbox.wld2pix(*gbox.pix2wld(133, 83)) == pytest.approx((133, 83), abs=1)

    assert gbox.zoom_to(50).shape.wh == (50, 47)

    assert gbox.zoom_out(0.5).zoom_out(2) == gbox

    assert len(set([gbox, gbox.center_pixel, gbox[:1, :2], gbox[:1, :2]])) == 3

    assert len(gbox.gcps()) == 24

    gbox_ = gbox.to_crs("epsg:3857")
    assert gbox_.shape == gbox.shape
    assert gbox_.crs == "epsg:3857"
    assert gbox_.wld2pix(*gbox_.pix2wld(0, 0)) == pytest.approx((0, 0), abs=1)
    assert gbox_.wld2pix(*gbox_.pix2wld(133, 83)) == pytest.approx((133, 83), abs=1)

    gbox_ = gbox.to_crs("utm")
    assert gbox_.shape == gbox.shape
    assert gbox_.crs is not None
    assert gbox_.crs.proj.utm_zone is not None
    assert gbox_.wld2pix(*gbox_.pix2wld(0, 0)) == pytest.approx((0, 0), abs=1)
    assert gbox_.wld2pix(*gbox_.pix2wld(133, 83)) == pytest.approx((133, 83), abs=1)

    assert gbox.pad(2).to_crs("epsg:6933").crs == "epsg:6933"

    p1, p2 = gbox.map_bounds()
    assert p1 == pytest.approx((-44.50301336231415, 109.39806656168265))
    assert p2 == pytest.approx((-9.47177497427409, 157.04711254391185))

    # map bounds without CRS, should still work, since data is in epsg:4326
    _mapping = GCPMapping(gbox._mapping._pix, gbox._mapping._wld, None)
    assert _mapping.crs is None
    gbox_ = GCPGeoBox(gbox.shape, _mapping)
    assert gbox_.crs is None
    p1, p2 = gbox_.map_bounds()
    assert p1 == pytest.approx((-44.50301336231415, 109.39806656168265))
    assert p2 == pytest.approx((-9.47177497427409, 157.04711254391185))

    assert gbox.anchor == AnchorEnum.FLOATING


def test_gcp_geobox_xr(au_gcp_geobox: GCPGeoBox):
    gbox = au_gcp_geobox
    xx = xr_zeros(gbox)
    _gbox = xx.odc.geobox
    assert _gbox.shape == gbox.shape
    assert _gbox.crs == gbox.crs
    assert (_gbox.extent ^ gbox.extent).is_empty

    # check the case when x/y coords are not populated
    yy = xx.drop_vars(xx.odc.spatial_dims)
    assert len(yy.coords) == 1
    _gbox = yy.odc.geobox

    assert _gbox.shape == gbox.shape
    assert _gbox.crs == gbox.crs
    assert (_gbox.extent ^ gbox.extent).is_empty

    # corrupt gcps
    yy.spatial_ref.attrs["gcps"] = "not even geojson"
    # should not throw, just return None
    assert yy.odc.uncached.geobox is None


def test_gcp_reproject(au_gcp_geobox: GCPGeoBox):
    # smoke-test only
    gbox = au_gcp_geobox
    xx = xr_zeros(gbox, time=["2020-02-20", "2021-01-21"], dtype="uint8", nodata=255)
    assert xx.ndim == 3
    assert xx.odc.nodata == 255
    assert (xx.odc.geobox.extent ^ gbox.extent).is_empty
    assert isinstance(xx.odc.geobox, GCPGeoBox)

    yy_gbox = xx.odc.geobox.approx.zoom_to(320)
    yy = xx.odc.reproject(yy_gbox)
    assert yy.odc.nodata == 255
    assert yy.odc.geobox == yy_gbox
    assert yy.ndim == 3
    assert yy.shape[0] == xx.shape[0]
    assert set(np.unique(yy.values).tolist()) == set([255, 0])


def test_gcp_mapping():
    gbox0 = GeoBox.from_bbox([0, 0, 20, 10], crs="epsg:4326", resolution=1)
    assert gbox0.shape.wh == (20, 10)
    assert gbox0.crs == "epsg:4326"

    px, py = gbox0.boundary(4).T
    pix = geom.multipoint(
        [(float(x), float(y)) for x, y in zip(px.tolist(), py.tolist())], None
    )
    wld = gbox0.project(pix)

    mapping = GCPMapping(pix, wld)
    assert mapping.crs == gbox0.crs
    _pix, _wld = mapping.points()
    assert _wld.crs == gbox0.crs
    assert _pix.crs is None

    assert (pix - _pix).is_empty
    assert (wld - _wld).is_empty

    assert mapping.approx[:6] == pytest.approx(gbox0.affine[:6], abs=1e-6)

    for pt in pix.geoms:
        # there and back with minimal loss of precision
        _pt = pt.transform(mapping.p2w).transform(mapping.w2p)
        assert _pt.coords[0] == pytest.approx(pt.coords[0], abs=1e-6)

    for pt in wld.geoms:
        # there and back with minimal loss of precision
        _pt = pt.transform(mapping.w2p).transform(mapping.p2w)
        assert _pt.coords[0] == pytest.approx(pt.coords[0], abs=1e-6)

    _m2 = GCPMapping(mapping._pix, wld)
    assert _m2.crs == mapping.crs
    assert _m2._pix is mapping._pix


@pytest.mark.parametrize("n", [3, 4, 8])
def test_gcp_few_points(n):
    gbox0 = GeoBox.from_bbox([0, 0, 20, 10], crs="epsg:4326", resolution=1)
    if n <= 4:
        corners = gbox0.boundary(2)[:n]
    else:
        corners = gbox0.boundary(n * 3 // 4)[:n:3]

    pix = geom.multipoint([(float(x), float(y)) for x, y in corners], None)
    wld = gbox0.project(pix)

    mapping = GCPMapping(pix, wld)
    gbox = GCPGeoBox(gbox0.shape, mapping)

    wld_ = gbox.project(pix)
    p1 = np.vstack([pt.coords for pt in wld.geoms])
    p2 = np.vstack([pt.coords for pt in wld_.geoms])
    np.testing.assert_array_almost_equal(p1, p2)



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2020 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
"""
Tools for dealing with ROIs (Regions of Interest).

In this context ROI is a 2d slice of an image. For example a top left corner of 10 pixels square
will have an ROI that can be constructed with :py:func:`numpy.s_` like this: ``s_[0:10, 0:10]``.
"""
import math
from collections import abc
from typing import List, Optional, Protocol, Sequence, Tuple, Union, overload

import numpy as np

from .math import align_down, align_up, edge_index
from .types import (
    ROI,
    Chunks2d,
    NdROI,
    NormalizedROI,
    NormalizedSlice,
    Shape2d,
    SomeIndex2d,
    SomeShape,
    SomeSlice,
    T,
    iyx_,
    shape_,
)

# This is numeric code, short names make sense in this context, so disabling
# "invalid name" checks for the whole file
# pylint: disable=invalid-name


class WindowFromSlice:
    """Translate numpy slices to rasterio window tuples."""

    def __getitem__(self, roi):
        if roi is None:
            return None

        if not isinstance(roi, abc.Sequence) or len(roi) != 2:
            raise ValueError("Need 2d roi")

        row, col = roi
        return (
            (0 if row.start is None else row.start, row.stop),
            (0 if col.start is None else col.start, col.stop),
        )


w_ = WindowFromSlice()


class RoiTiles(Protocol):
    """
    Abstraction for 2d slice/shape/chunks lookup.
    """

    def __getitem__(self, idx: Union[SomeIndex2d, ROI]) -> Tuple[slice, slice]: ...

    def crop(self, roi: ROI) -> "RoiTiles": ...

    def tile_shape(self, idx: SomeIndex2d) -> Shape2d: ...

    @property
    def shape(self) -> Shape2d: ...

    @property
    def base(self) -> Shape2d: ...

    @property
    def chunks(self) -> Chunks2d: ...

    def locate(self, pix: SomeIndex2d) -> Tuple[int, int]: ...

    def __dask_tokenize__(self): ...


def norm_slice_2d(
    idx: Union[SomeIndex2d, ROI], shape: Tuple[int, int]
) -> NormalizedROI:
    if isinstance(idx, tuple):
        return roi_normalise(idx, shape)
    return roi_normalise(iyx_(idx).yx, shape)


def _fmt_shape(shape):
    n1, n2 = shape.yx
    if max(n1, n2) > 10_000:
        return f"{n1:_d}x{n2:_d}"
    return f"{n1:d}x{n2:d}"


def clip_tiles(
    tiles: RoiTiles, selection: Sequence[Tuple[int, int]]
) -> Tuple["RoiTiles", Tuple[slice, slice], List[Tuple[int, int]]]:
    """
    Compute cropped version of tiles from a selection of tile indexes.

    :return: Cropped RoiTiles object, crop that was applied and tile indexes
             changed to the cropped address space.
    """
    ii = np.asarray(selection)
    y1, x1 = ii.min(axis=0).tolist()
    y2, x2 = ii.max(axis=0).tolist()
    roi = np.s_[y1 : y2 + 1, x1 : x2 + 1]
    sel_new = [(y - y1, x - x1) for y, x in selection]
    return tiles.crop(roi), roi, sel_new


class Tiles:
    """
    Partition box into tiles.

    Turns ``row, col`` index into a 2d ROI of the original box.
    """

    def __init__(self, base_shape: SomeShape, tile_shape: SomeShape) -> None:
        tile_shape = shape_(tile_shape)
        base_shape = shape_(base_shape)
        self._tile_shape = tile_shape
        self._base_shape = base_shape
        ny, nx = (
            int(math.ceil(float(N) / n)) for N, n in zip(base_shape.yx, tile_shape.yx)
        )
        self._shape = shape_((ny, nx))

    def crop(self, roi: ROI) -> "Tiles":
        base_shape = roi_shape(self[roi])
        return Tiles(base_shape, self._tile_shape)

    def __getitem__(self, idx: Union[SomeIndex2d, ROI]) -> Tuple[slice, slice]:
        def _slice(i: NormalizedSlice, N: int, n: int) -> slice:
            _in = i.start * n
            _out = i.stop * n

            if 0 <= _in < N and _out < N + n:
                return slice(_in, min(_out, N))
            raise IndexError(f"Index {idx} is out of range")

        idx = norm_slice_2d(idx, self._shape.yx)

        ir, ic = (
            _slice(i, N, n)
            for i, N, n in zip(idx, self._base_shape.yx, self._tile_shape.yx)
        )
        return (ir, ic)

    def tile_shape(self, idx: SomeIndex2d) -> Shape2d:
        """
        Query shape for a given tile.

        :param idx: ``(row, col)`` chunk index, supports numpy style indexing.
        :returns: shape of a tile (edge tiles might be smaller)
        :raises: :py:class:`IndexError` when index is outside of ``[(0,0) -> .shape)``.
        """
        idx = iyx_(idx)

        def _sz(i: int, n: int, tile_sz: int, total_sz: int) -> int:
            if i < 0:  # numpy style index from the right
                i = n + i
            if 0 <= i < n - 1:  # not edge tile
                return tile_sz
            if i == n - 1:  # edge tile
                return total_sz - (i * tile_sz)
            # out of index case
            raise IndexError(f"Index {idx} is out of range")

        ny, nx = map(
            _sz, idx.yx, self._shape.yx, self._tile_shape.yx, self._base_shape.yx
        )
        return shape_((ny, nx))

    @property
    def shape(self) -> Shape2d:
        """Number of tiles along each dimension."""
        return self._shape

    @property
    def base(self) -> Shape2d:
        """Base shape."""
        return self._base_shape

    @property
    def chunks(self) -> Chunks2d:
        """Dask compatible chunk rerpesentation."""
        NY, NX = self.shape.yx
        ny, nx = self.tile_shape((0, 0)).yx
        ny_, nx_ = self.tile_shape((NY - 1, NX - 1))

        return (
            (ny,) * (NY - 1) + (ny_,),
            (nx,) * (NX - 1) + (nx_,),
        )

    def locate(self, pix: SomeIndex2d) -> Tuple[int, int]:
        """Tile index from pixel coordinate."""
        NY, NX = self._base_shape.yx
        y, x = iyx_(pix).yx
        if y < 0 or y >= NY or x < 0 or x >= NX:
            raise IndexError()
        ny, nx = self.tile_shape((0, 0)).yx
        return (y // ny, x // nx)

    def __dask_tokenize__(self):
        return (
            "odc.geo.roi.Tiles",
            *self._shape,
            *self._tile_shape,
        )

    def __str__(self) -> str:
        b1, b2, b3 = map(_fmt_shape, [self._shape, self._tile_shape, self._base_shape])
        return f"Tiles: {b1}|{b2}px => {b3}px"

    def __repr__(self) -> str:
        return self.__str__()

    def __eq__(self, __value: object) -> bool:
        if __value is self:
            return True
        if not isinstance(__value, Tiles):
            return False
        return (
            self._base_shape == __value._base_shape
            and self._tile_shape == __value._tile_shape
        )


class VariableSizedTiles:
    """
    Partition box into tiles of varying sizes.

    Turns ``row, col`` index into a 2d ROI of the original box.
    """

    __slots__ = ("_offsets",)

    def __init__(self, chunks: Chunks2d) -> None:
        self._offsets = tuple(
            np.asarray([0, *idx], dtype="int32").cumsum(dtype="int32") for idx in chunks
        )

    def crop(self, roi: ROI) -> "VariableSizedTiles":
        roi = roi_normalise(roi, self.shape.yx)
        y, x = (ch[s.start : s.stop] for ch, s in zip(self.chunks, roi))
        return VariableSizedTiles((y, x))

    def __getitem__(self, idx: Union[SomeIndex2d, ROI]) -> Tuple[slice, slice]:
        idx = norm_slice_2d(idx, self.shape.yx)
        y, x = (
            slice(int(a[i.start]), int(a[i.stop])) for a, i in zip(self._offsets, idx)
        )
        return (y, x)

    def tile_shape(self, idx: SomeIndex2d) -> Shape2d:
        """
        Query shape for a given tile.

        :param idx: ``(row, col)`` chunk index
        :returns: shape of a tile
        :raises: :py:class:`IndexError` when index is outside of ``[(0,0) -> .shape)``.
        """
        idx = iyx_(idx)
        ny, nx = (int(a[i + 1]) - int(a[i]) for a, i in zip(self._offsets, idx.yx))
        return Shape2d(x=nx, y=ny)

    @property
    def shape(self) -> Shape2d:
        """Number of tiles along each dimension."""
        ny, nx = (len(idx) - 1 for idx in self._offsets)
        return Shape2d(x=nx, y=ny)

    @property
    def base(self) -> Shape2d:
        """Base shape."""
        ny, nx = (int(idx[-1]) for idx in self._offsets)
        return Shape2d(x=nx, y=ny)

    @property
    def chunks(self) -> Tuple[Tuple[int, ...], Tuple[int, ...]]:
        """Dask compatible chunk rerpesentation."""
        y, x = (tuple(map(int, np.diff(idx))) for idx in self._offsets)
        return (y, x)

    def locate(self, pix: SomeIndex2d) -> Tuple[int, int]:
        """Tile index from pixel coordinate."""
        NY, NX = self.base.yx
        y, x = iyx_(pix).yx
        if y < 0 or y >= NY or x < 0 or x >= NX:
            raise IndexError()
        y, x = (
            # 1: because offsets start from 0, n1, n1+n2, ...
            int(np.searchsorted(bins[1:], pix, "right"))
            for bins, pix in zip(self._offsets, (y, x))
        )
        return (y, x)

    def __dask_tokenize__(self):
        return (
            "odc.geo.roi.VariableSizedTiles",
            *self._offsets,
        )

    def __str__(self) -> str:
        b1, b2, b3 = map(_fmt_shape, [self.shape, self.tile_shape((0, 0)), self.base])
        return f"Tiles: {b1}|chunked {b2}px => {b3}px"

    def __repr__(self) -> str:
        return self.__str__()

    def __eq__(self, __value: object) -> bool:
        if __value is self:
            return True
        if not isinstance(__value, VariableSizedTiles):
            return False

        for a, b in zip(self._offsets, __value._offsets):
            if a.shape != b.shape:
                return False
            if (a != b).any():
                return False
        return True


def roi_tiles(shape: SomeShape, how: Union[SomeShape, Chunks2d]) -> RoiTiles:
    if isinstance(how, (tuple, list)) and isinstance(how[0], (tuple, list)):
        y, x = (tuple(i) for i in how)
        return VariableSizedTiles((y, x))
    return Tiles(shape, how)


def polygon_path(
    x: np.ndarray,
    y: Optional[np.ndarray] = None,
    closed: bool = True,
) -> np.ndarray:
    """
    Points along axis aligned polygon.

    A little bit like :py:func:`numpy.meshgrid`, except returns only boundary points and limited to
    a 2d case only.

    .. rubric:: Examples

    .. code-block::

       [0,1] - unit square
       [0,1], [0,1] => [[0, 1, 1, 0, 0],
                        [0, 0, 1, 1, 0]])

       # three points per X, two point per Y side
       [0,1,2], [7,9] => [[0, 1, 2, 2, 1, 0, 0],
                          [7, 7, 7, 9, 9, 9, 7]]

    :returns:
      A ``2xN`` array, ``x, y = polygon_path(...)``

    """

    if not isinstance(x, np.ndarray):
        x = np.asarray(x)

    if y is None:
        y = x

    if not isinstance(y, np.ndarray):
        y = np.asarray(y)

    shape = (len(y), len(x))
    iy, ix = np.asarray(list(edge_index(shape, closed=closed)), dtype="uint32").T
    return np.vstack([x[np.newaxis, ix], y[np.newaxis, iy]])


def roi_boundary(roi: NormalizedROI, pts_per_side: int = 2) -> np.ndarray:
    """
    Get boundary points from a 2d roi.

    roi needs to be in the normalised form, i.e. no open-ended start/stop,

    :returns:
      ``Nx2 <float64>`` array of ``X,Y`` points on the perimeter of the envelope defined by ``roi``

    .. seealso:: :py:func:`~odc.geo.roi.roi_normalise`
    """
    yy, xx = roi
    xx = np.linspace(xx.start, xx.stop, pts_per_side, dtype="float64")
    yy = np.linspace(yy.start, yy.stop, pts_per_side, dtype="float64")

    return polygon_path(xx, yy, closed=False).T


def scaled_down_roi(roi: NormalizedROI, scale: int) -> NormalizedROI:
    """
    Compute ROI for a scaled down image.

    Given a crop region of the original image compute equivalent crop in the overview image.

    :param roi: ROI in the original image
    :param scale: integer scale to get scaled down image
    :return: ROI in the scaled down image
    """
    s1, s2 = (slice(s.start // scale, align_up(s.stop, scale) // scale) for s in roi)
    return (s1, s2)


def scaled_up_roi(
    roi: NormalizedROI, scale: int, shape: Optional[SomeShape] = None
) -> NormalizedROI:
    """
    Compute ROI for a scaled up image.

    Given a crop region in the original image compute equivalent crop in the upsampled image.

    :param roi: ROI in the original image
    :param scale: integer scale to get scaled up image
    :param shape: Clamp to that shape is supplied
    :return: ROI in the scaled up image
    """
    s1, s2 = (slice(s.start * scale, s.stop * scale) for s in roi)
    if shape is not None:
        s1, s2 = (
            slice(min(dim, s.start), min(dim, s.stop))
            for s, dim in zip([s1, s2], shape_(shape))
        )
    return (s1, s2)


def scaled_down_shape(shape: Tuple[int, ...], scale: int) -> Tuple[int, ...]:
    """
    Compute shape of the overview image.

    :param shape: Original shape
    :param scale: Shrink factor
    :return: shape of the overview image
    """
    return tuple(align_up(s, scale) // scale for s in shape)


# fmt: off
@overload
def roi_shape(roi: ROI) -> Tuple[int, int]: ...
@overload
def roi_shape(roi: NdROI) -> Tuple[int, ...]: ...
# fmt: on


def roi_shape(roi: NdROI) -> Tuple[int, ...]:
    """
    Shape of an array after cropping with ``roi``.

    Same as ``xx[roi].shape``.
    """

    def slice_dim(s: SomeSlice) -> int:
        if isinstance(s, int):
            return 1
        _out = s.stop
        if _out is None:
            raise ValueError(
                "Can't determine shape of the slice with open right-hand side."
            )
        if s.start is None:
            return _out
        return _out - s.start

    if not isinstance(roi, tuple):
        roi = (roi,)

    return tuple(slice_dim(s) for s in roi)


def roi_is_empty(roi: NdROI) -> bool:
    """
    Check if ROI is "empty".

    ROI is empty if any dimension is 0 elements wide.
    """
    return any(d <= 0 for d in roi_shape(roi))


def roi_is_full(roi: NdROI, shape: Union[int, Tuple[int, ...]]) -> bool:
    """
    Check if ROI covers the entire region.

    :returns: ``True`` if ``roi`` covers region from ``(0,..) -> shape``
    :returns: ``False`` if ``roi`` actually crops an image
    """

    def slice_full(s: SomeSlice, n: int) -> bool:
        if isinstance(s, int):
            return n == 1
        return s.start in (0, None) and s.stop in (n, None)

    if not isinstance(roi, tuple):
        roi = (roi,)

    if not isinstance(shape, tuple):
        shape = (shape,)

    return all(slice_full(s, n) for s, n in zip(roi, shape))


def _fill_if_none(x: Optional[T], val_if_none: T) -> T:
    return val_if_none if x is None else x


def _norm_slice_or_error(s: SomeSlice) -> NormalizedSlice:
    if isinstance(s, int):
        start = s
        stop = s + 1
        step = None
    else:
        start = _fill_if_none(s.start, 0)

        if s.stop is None:
            raise ValueError("Can't process open ended slice")

        stop = s.stop
        step = s.step

    if stop < 0 or start < 0:
        raise ValueError("Can't process negative offset slice")

    return slice(start, stop, step)


def _norm_slice(s: SomeSlice, n: int) -> NormalizedSlice:
    if isinstance(s, int):
        if s < 0:
            s = n + s
        return slice(s, s + 1)
    start = _fill_if_none(s.start, 0)
    stop = _fill_if_none(s.stop, n)
    start, stop = (x if x >= 0 else n + x for x in (start, stop))
    return slice(start, stop, s.step)


def slice_intersect3(a: SomeSlice, b: SomeSlice) -> Tuple[slice, slice, slice]:
    """
    Compute overlap 3 way.

    Compute part of a that overlaps b, part of b that overlaps a and the region
    of the original region covered by the overlap.

    :returns: ``a', b', ab'``, such that ``X[a][a'] == X[b][b'] == X[ab']``
    """
    a = _norm_slice_or_error(a)
    b = _norm_slice_or_error(b)
    na = a.stop - a.start
    nb = b.stop - b.start

    if a.stop < b.start:
        return slice(na, na), slice(0, 0), slice(a.stop, a.stop)
    if a.start > b.stop:
        return slice(0, 0), slice(nb, nb), slice(a.start, a.start)

    _in = max(a.start, b.start)
    _out = min(a.stop, b.stop)

    return (
        slice(_in - a.start, _out - a.start),
        slice(_in - b.start, _out - b.start),
        slice(_in, _out),
    )


def roi_intersect3(
    a: Tuple[SomeSlice, ...], b: Tuple[SomeSlice, ...]
) -> Tuple[Tuple[slice, ...], Tuple[slice, ...], Tuple[slice, ...]]:
    """
    Compute overlap 3 way.

    Compute part of a that overlaps b, part of b that overlaps a and the region
    of the original region covered by the overlap.

    :returns: ``a', b', ab'``, such that ``X[a][a'] == X[b][b'] == X[ab']``
    """
    assert len(a) == len(b)
    aa, bb, cc = zip(*[slice_intersect3(a_, b_) for a_, b_ in zip(a, b)])
    return aa, bb, cc


# fmt: off
@overload
def roi_normalise(roi: SomeSlice, shape: Union[int, Tuple[int]]) -> NormalizedSlice: ...
@overload
def roi_normalise(roi: ROI, shape: Tuple[int, int]) -> NormalizedROI: ...
@overload
def roi_normalise(roi: Tuple[SomeSlice, ...], shape: Tuple[int, ...]
) -> Tuple[NormalizedSlice, ...]: ...
# fmt: on


def roi_normalise(
    roi: NdROI, shape: Union[int, Tuple[int, ...]]
) -> Union[NormalizedSlice, Tuple[NormalizedSlice, ...]]:
    """
    Normalise ROI.

    Fill in missing ``.start/.stop``, also deal with negative values, which are treated as offsets
    from the end.

    ``.step`` parameter is left unchanged.

    .. rubric:: Example

    .. code-block::

       np.s_[:3, 4:  ], (10, 20) => np.s_[0:3, 4:20]
       np.s_[:3,  :-3], (10, 20) => np.s_[0:3, 0:17]

    """
    if not isinstance(roi, abc.Sequence):
        if isinstance(shape, abc.Sequence):
            (shape,) = shape
        return _norm_slice(roi, shape)

    if not isinstance(shape, abc.Sequence):
        shape = (shape,)

    return tuple(_norm_slice(s, n) for s, n in zip(roi, shape))


# fmt: off
@overload
def roi_pad(roi: SomeSlice, pad: int, shape: int) -> NormalizedSlice: ...
@overload
def roi_pad(roi: Tuple[SomeSlice, ...], pad: int, shape: Tuple[int, ...]) -> Tuple[NormalizedSlice, ...]: ...
# fmt: on


def roi_pad(
    roi: NdROI, pad: int, shape: Union[int, Tuple[int, ...]]
) -> Union[NormalizedSlice, Tuple[NormalizedSlice, ...]]:
    """
    Pad ROI on each side, with clamping.

    Returned ROI is guaranteed to be within ``(0,..) -> shape``.
    """

    def pad_slice(s: SomeSlice, n: int) -> NormalizedSlice:
        s = _norm_slice(s, n)
        return slice(max(0, s.start - pad), min(n, s.stop + pad))

    if not isinstance(roi, abc.Sequence):
        if isinstance(shape, abc.Sequence):
            (shape,) = shape
        return pad_slice(roi, shape)

    if not isinstance(shape, abc.Sequence):
        shape = (shape,)

    return tuple(pad_slice(s, n) for s, n in zip(roi, shape))


# fmt: off
@overload
def roi_intersect(a: SomeSlice, b: SomeSlice) -> NormalizedSlice: ...
@overload
def roi_intersect(a: Tuple[SomeSlice, ...], b: Tuple[SomeSlice, ...]) -> Tuple[NormalizedSlice, ...]: ...
# fmt: on


def roi_intersect(
    a: NdROI, b: NdROI
) -> Union[NormalizedSlice, Tuple[NormalizedSlice, ...]]:
    """
    Compute intersection of two ROIs.

    .. rubric:: Examples

    .. code-block::

       s_[1:30], s_[20:40] => s_[20:30]
       s_[1:10], s_[20:40] => s_[10:10]

       # works for N dimensions
       s_[1:10, 11:21], s_[8:12, 10:30] => s_[8:10, 11:21]

    """

    def slice_intersect(a: SomeSlice, b: SomeSlice) -> NormalizedSlice:
        a = _norm_slice_or_error(a)
        b = _norm_slice_or_error(b)

        if a.stop < b.start:
            return slice(a.stop, a.stop)
        if a.start > b.stop:
            return slice(a.start, a.start)

        _in = max(a.start, b.start)
        _out = min(a.stop, b.stop)

        return slice(_in, _out)

    if not isinstance(a, abc.Sequence):
        if isinstance(b, abc.Sequence):
            (b,) = b
        return slice_intersect(a, b)

    if not isinstance(b, abc.Sequence):
        b = (b,)

    return tuple(slice_intersect(sa, sb) for sa, sb in zip(a, b))


# fmt: off
@overload
def roi_center(roi: SomeSlice) -> float: ...
@overload
def roi_center(roi: Tuple[SomeSlice, ...]) -> Tuple[float, ...]: ...
# fmt: on


def roi_center(roi: NdROI) -> Union[float, Tuple[float, ...]]:
    """Return center point of an ``roi``."""

    def slice_center(s: SomeSlice) -> float:
        s = _norm_slice_or_error(s)
        return (s.start + s.stop) * 0.5

    if not isinstance(roi, abc.Sequence):
        return slice_center(roi)

    return tuple(slice_center(s) for s in roi)


def roi_from_points(
    xy: np.ndarray,
    shape: SomeShape,
    padding: int = 0,
    align: Optional[int] = None,
) -> NormalizedROI:
    """
    Build ROI from sample points.

    Compute envelope around a bunch of points and return it as an ROI (tuple of
    row/col slices)

    Returned roi is clipped ``(0,0) --> shape``, so it won't stick outside of the
    valid region.
    """
    shape = shape_(shape)

    def to_roi(*args):
        return tuple(slice(int(v[0]), int(v[1])) for v in args)

    assert len(shape) == 2
    assert xy.ndim == 2 and xy.shape[1] == 2

    # keep finite points only
    #  if any points are not finite remove them
    ok_mask = np.isfinite(xy)
    if not ok_mask.all():
        keep = ok_mask.T[0] * ok_mask.T[1]
        xy = xy[keep, :]

    if xy.shape[0] == 0:
        return np.s_[0:0, 0:0]

    ny, nx = shape

    _in = np.floor(xy.min(axis=0)).astype("int32") - padding
    _out = np.ceil(xy.max(axis=0)).astype("int32") + padding

    if align is not None:
        _in = align_down(_in, align)
        _out = align_up(_out, align)

    xx = np.asarray([_in[0], _out[0]])
    yy = np.asarray([_in[1], _out[1]])

    xx = np.clip(xx, 0, nx, out=xx)
    yy = np.clip(yy, 0, ny, out=yy)

    return to_roi(yy, xx)



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2020 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
import numpy as np
import pytest
import xarray as xr

from odc.geo import geom
from odc.geo._interop import is_dask_collection
from odc.geo._xr_interop import _extract_geo_transform
from odc.geo.data import ocean_geom
from odc.geo.geobox import GeoBox
from odc.geo.testutils import approx_equal_geobox, epsg3577, mkA, purge_crs_info
from odc.geo.types import ROI, resxy_
from odc.geo.xr import (
    ODCExtensionDa,
    ODCExtensionDs,
    rasterize,
    register_geobox,
    spatial_dims,
    wrap_xr,
    xr_coords,
    xr_reproject,
    xr_zeros,
)

# pylint: disable=redefined-outer-name,import-outside-toplevel,protected-access
TEST_GEOBOXES_SMALL_AXIS_ALIGNED = [
    GeoBox.from_bbox((-10, -2, 5, 4), "epsg:4326", tight=True, resolution=0.2),
    GeoBox.from_bbox((-10, -2, 5, 4), "epsg:3857", tight=True, resolution=1),
    GeoBox.from_bbox((-10, -2, 5, 4), "epsg:3857", tight=True, resolution=resxy_(1, 2)),
]

NaN = float("nan")


@pytest.fixture
def geobox_epsg4326():
    _box = geom.box(-20, -10, 20, 10, "epsg:4326")
    yield GeoBox.from_geopolygon(_box, 5)


@pytest.fixture
def xx_chunks():
    yield None


@pytest.fixture
def xx_time():
    yield None


@pytest.fixture
def xx_epsg4326(geobox_epsg4326: GeoBox, xx_time, xx_chunks):
    if xx_time is not None and xx_chunks is not None:
        if len(xx_chunks) < 3:
            xx_chunks = (-1, *xx_chunks)
    yield xr_zeros(geobox_epsg4326, dtype="uint16", chunks=xx_chunks, time=xx_time)


def test_geobox_xr_coords():
    A = mkA(0, scale=(10, -10), translation=(-48800, -2983006))

    w, h = 512, 256
    _shape = h, w
    gbox = GeoBox(_shape, A, epsg3577)

    cc = xr_coords(gbox, crs_coord_name=None)
    assert list(cc) == ["y", "x"]
    assert cc["y"].shape == (gbox.shape[0],)
    assert cc["x"].shape == (gbox.shape[1],)
    assert "crs" in cc["y"].attrs
    assert "crs" in cc["x"].attrs

    cc = xr_coords(gbox)
    assert list(cc) == ["y", "x", "spatial_ref"]
    assert cc["spatial_ref"].shape == ()
    assert cc["spatial_ref"].attrs["spatial_ref"] == gbox.crs.wkt
    assert isinstance(cc["spatial_ref"].attrs["grid_mapping_name"], str)
    assert isinstance(cc["spatial_ref"].attrs["GeoTransform"], str)

    # crs_coord_name should be default "spatial_ref"
    assert list(xr_coords(gbox)) == list(xr_coords(gbox, crs_coord_name="spatial_ref"))

    cc = xr_coords(gbox, crs_coord_name="Albers")
    assert list(cc) == ["y", "x", "Albers"]

    # non-axis aligned case
    _gbox = gbox.rotate(33)
    assert _gbox.axis_aligned is False
    cc = xr_coords(_gbox)
    assert list(cc) == ["y", "x", "spatial_ref"]
    assert cc["spatial_ref"].shape == ()
    assert cc["spatial_ref"].attrs["spatial_ref"] == gbox.crs.wkt
    assert isinstance(cc["spatial_ref"].attrs["grid_mapping_name"], str)
    assert isinstance(cc["spatial_ref"].attrs["GeoTransform"], str)
    assert xr_zeros(_gbox, dtype="uint16").odc.geobox == _gbox

    # geographic CRS
    A = mkA(0, scale=(0.1, -0.1), translation=(10, 30))
    gbox = GeoBox(_shape, A, "epsg:4326")

    cc = xr_coords(gbox)
    assert list(cc) == ["latitude", "longitude", "spatial_ref"]
    assert cc["spatial_ref"].shape == ()
    assert cc["spatial_ref"].attrs["spatial_ref"] == gbox.crs.wkt
    assert isinstance(cc["spatial_ref"].attrs["grid_mapping_name"], str)


def test_xr_zeros(geobox_epsg4326: GeoBox):
    # missing CRS for GeoBox
    gbox = geobox_epsg4326
    xx = xr_zeros(gbox, dtype="uint16")
    assert "spatial_ref" in xx.coords
    assert xx.encoding["grid_mapping"] == "spatial_ref"
    assert (xx.values == 0).all()
    assert gbox.crs is not None
    assert xx.spatial_ref.attrs["spatial_ref"] == gbox.crs.wkt
    assert xx.spatial_ref.attrs["grid_mapping_name"] == "latitude_longitude"
    assert xx.odc.ydim == 0
    assert xx.odc.xdim == 1

    # check custom name for crs coordinate
    xx = xr_zeros(gbox, dtype="uint16", crs_coord_name="_crs")
    assert "_crs" in xx.coords
    assert xx.encoding["grid_mapping"] == "_crs"
    assert (xx.values == 0).all()

    # rotated gebox
    gbox = geobox_epsg4326.center_pixel.rotate(-45).pad_wh(10, 20)
    xx = xr_zeros(gbox, dtype="uint16")
    assert xx.odc.geobox == gbox
    assert xx[3:, 1:].odc.geobox.affine[:6] == pytest.approx(gbox[3:, 1:].affine[:6])
    assert xx[5:, 7:].odc.geobox.affine[:6] == pytest.approx(gbox[5:, 7:].affine[:6])

    # dask version
    gbox = geobox_epsg4326.zoom_to((100, 200))
    xx = xr_zeros(gbox, dtype="uint16", chunks=(10, 20))
    assert is_dask_collection(xx) is True
    assert xx.dtype == "uint16"
    assert xx.data.chunksize == (10, 20)
    assert xx.shape == gbox.shape

    # time axis
    gbox = geobox_epsg4326.zoom_to((100, 200))
    xx = xr_zeros(
        gbox, dtype="uint16", chunks=(1, 10, 20), time=["2020-01-01", "2020-01-02"]
    )
    assert is_dask_collection(xx) is True
    assert xx.dtype == "uint16"
    assert xx.data.chunksize == (1, 10, 20)
    assert xx.shape == (2, *gbox.shape)
    assert xx.odc.ydim == 1
    assert xx.odc.xdim == 2


def test_purge_crs_info(xx_epsg4326: xr.DataArray):
    xx = xx_epsg4326
    assert xx.odc.crs is not None
    assert purge_crs_info(xx).odc.crs is None

    xx = purge_crs_info(xx_epsg4326)
    assert xx.odc.crs is None
    assert xx.encoding == {}


def test_set_nodata(xx_epsg4326: xr.DataArray):
    xx = xx_epsg4326

    assert xx.odc.nodata is None
    assert xx.attrs == {}

    for new_nodata in [0, 999, 3]:
        xx.odc.nodata = new_nodata
        assert xx.odc.nodata == new_nodata
        assert "_FillValue" in xx.encoding
        assert xx.encoding["_FillValue"] == new_nodata
        assert xx.attrs["nodata"] == new_nodata

        xx.odc.nodata = None
        assert xx.odc.nodata is None
        assert "_FillValue" not in xx.encoding
        assert xx.attrs == {}


def test_nodata_nan(xx_epsg4326: xr.DataArray):
    xx = xx_epsg4326.astype("float32")
    assert isinstance(xx.odc, ODCExtensionDa)

    assert xx.odc.nodata is None
    assert xx.dtype.kind == "f"

    # set to NaN from nothing
    xx.odc.nodata = NaN
    assert isinstance(xx.odc.nodata, float)
    assert np.isnan(xx.odc.nodata)
    assert np.isnan(xx.encoding["_FillValue"])
    assert np.isnan(xx.attrs["nodata"])

    xx.odc.nodata = -9999
    assert xx.odc.nodata == -9999

    # set to NaN from non-NaN, and use a string
    xx.odc.nodata = "nan"  # type: ignore
    assert isinstance(xx.odc.nodata, float)
    assert np.isnan(xx.odc.nodata)
    assert np.isnan(xx.encoding["_FillValue"])
    assert np.isnan(xx.attrs["nodata"])


def test_odc_extension(xx_epsg4326: xr.DataArray, geobox_epsg4326: GeoBox):
    gbox = geobox_epsg4326
    xx = xx_epsg4326

    assert gbox.crs is not None
    assert "spatial_ref" in xx.coords
    assert xx.encoding["grid_mapping"] == "spatial_ref"
    assert xx.odc.geobox == gbox
    assert (xx.values == 0).all()
    assert xx.odc.crs == "epsg:4326"
    assert xx.odc.transform == gbox.transform
    assert xx.odc.spatial_dims == ("latitude", "longitude")
    assert xx.spatial_ref.attrs["spatial_ref"] == gbox.crs.wkt
    assert xx.spatial_ref.attrs["grid_mapping_name"] == "latitude_longitude"
    assert xx.odc.uncached.transform == xx.odc.transform
    assert xx.odc.output_geobox("epsg:3857").crs == "epsg:3857"
    assert xx.odc.map_bounds() == gbox.map_bounds()
    assert xx.odc.output_geobox("utm").crs.epsg is not None
    assert xx.odc.aspect == gbox.aspect
    assert xx.odc.grid_mapping == xx.encoding["grid_mapping"]
    assert xx.odc.crs_coord == xx.spatial_ref
    assert xx.odc.crs_coord.attrs == xx.spatial_ref.attrs

    # this drops encoding/attributes, but crs/geobox should remain the same
    _xx = xx * 10.0
    assert _xx.encoding == {}
    assert _xx.odc.crs == xx.odc.crs
    assert _xx.odc.geobox == xx.odc.geobox

    # test non-standard coordinate names
    _xx = xx.rename(longitude="XX", latitude="YY")
    assert _xx.dims == ("YY", "XX")
    assert _xx.odc.spatial_dims == ("YY", "XX")
    assert _xx.odc.crs == gbox.crs

    # 1-d xarrays should report None for everything
    assert _xx.XX.odc.spatial_dims is None
    assert _xx.XX.odc.crs is None
    assert _xx.XX.odc.transform is None
    assert _xx.XX.odc.geobox is None

    # when geobox is none output_geobox should fail
    with pytest.raises(ValueError):
        _ = _xx.XX.odc.output_geobox("epsg:4326")

    # no geobox - no map bounds
    with pytest.raises(ValueError):
        _ = _xx.XX.odc.map_bounds()

    # no spatial_dims - no ydim/xdim
    with pytest.raises(ValueError):
        _ = _xx.XX.odc.ydim
    with pytest.raises(ValueError):
        _ = _xx.XX.odc.xdim


def test_odc_extension_ds(xx_epsg4326: xr.DataArray, geobox_epsg4326: GeoBox):
    gbox = geobox_epsg4326
    xx = xx_epsg4326.to_dataset(name="band")

    assert gbox.crs is not None
    assert "spatial_ref" in xx.coords
    assert xx.band.encoding["grid_mapping"] == "spatial_ref"
    assert xx.odc.geobox == gbox
    assert xx.odc.crs == "epsg:4326"
    assert xx.odc.transform == gbox.transform
    assert xx.odc.spatial_dims == ("latitude", "longitude")
    assert xx.spatial_ref.attrs["spatial_ref"] == gbox.crs.wkt
    assert xx.spatial_ref.attrs["grid_mapping_name"] == "latitude_longitude"
    assert xx.odc.uncached.transform == xx.odc.transform

    # this drops encoding/attributes, but crs/geobox should remain the same
    _xx = xx * 10.0
    assert _xx.band.encoding == {}
    assert _xx.odc.crs == xx.odc.crs
    assert _xx.odc.geobox == xx.odc.geobox

    # test non-standard coordinate names
    _xx = xx.rename(longitude="XX", latitude="YY")
    assert _xx.band.dims == ("YY", "XX")
    assert _xx.odc.spatial_dims == ("YY", "XX")
    assert _xx.odc.crs == gbox.crs

    # 1-d xarrays should report None for everything
    assert _xx.XX.odc.spatial_dims is None
    assert _xx.XX.odc.crs is None
    assert _xx.XX.odc.transform is None
    assert _xx.XX.odc.geobox is None


def test_assign_crs(xx_epsg4326: xr.DataArray):
    xx = purge_crs_info(xx_epsg4326)
    assert xx.odc.crs is None
    assert xx.odc.grid_mapping is None
    assert xx.odc.crs_coord is None
    yy = xx.odc.assign_crs("epsg:4326")
    assert xx.odc.uncached.crs is None
    assert yy.odc.crs == "epsg:4326"

    # non-cf complaint CRS
    yy = xx.odc.assign_crs("epsg:3857")
    assert yy.spatial_ref.attrs.get("grid_mapping_name") is None
    assert yy.odc.crs == "epsg:3857"


def test_assign_crs_ds(xx_epsg4326: xr.DataArray):
    xx = purge_crs_info(xx_epsg4326).to_dataset(name="band")
    assert xx.odc.crs is None
    yy = xx.odc.assign_crs("epsg:4326")
    assert xx.odc.uncached.crs is None
    assert yy.odc.crs == "epsg:4326"
    assert yy.band.odc.crs == "epsg:4326"
    assert yy.band.encoding["grid_mapping"] == "spatial_ref"

    # non-cf complaint CRS
    yy = xx.odc.assign_crs("epsg:3857")
    assert yy.spatial_ref.attrs.get("grid_mapping_name") is None
    assert yy.odc.crs == "epsg:3857"


def test_corrupt_inputs(xx_epsg4326: xr.DataArray):
    xx = xx_epsg4326.copy()
    assert xx.odc.crs == "epsg:4326"

    # incorrect grid_mapping "pointer"
    xx.encoding["grid_mapping"] = "bad-ref"
    with pytest.warns(UserWarning):
        # it will still get it from attributes
        assert xx.odc.uncached.crs == "epsg:4326"

    # missing spatial_ref, but have crs_wkt
    xx = xx_epsg4326.copy()
    xx.spatial_ref.attrs.pop("spatial_ref")
    assert xx.odc.crs == "epsg:4326"

    # missing wkt text
    xx = xx_epsg4326.copy()
    xx.spatial_ref.attrs.pop("spatial_ref")
    xx.spatial_ref.attrs.pop("crs_wkt")
    assert xx.odc.crs is None

    # bad CRS string
    xx = xx_epsg4326.copy()
    xx.spatial_ref.attrs["spatial_ref"] = "this is not a CRS!!!"
    xx.spatial_ref.attrs["crs_wkt"] = "this is not a CRS!!!"
    assert xx.odc.crs is None

    # duplicated crs coords and no grid_mapping pointer
    xx = xx_epsg4326.copy()
    xx = xx.assign_coords(_crs2=xx.spatial_ref)
    xx.encoding.pop("grid_mapping")
    with pytest.warns(UserWarning):
        assert xx.odc.crs == "epsg:4326"

    # attribute based search with bad CRS string
    xx = purge_crs_info(xx_epsg4326)
    assert xx.odc.crs is None
    xx = xx.assign_attrs(crs="!some invalid CRS string!!")
    with pytest.warns(UserWarning):
        assert xx.odc.crs is None

    # attribute based search with bad CRS type
    xx = purge_crs_info(xx_epsg4326)
    assert xx.odc.crs is None
    xx = xx.assign_attrs(crs=[])
    with pytest.warns(UserWarning):
        assert xx.odc.crs is None

    # attribute based search with several different CRS candidates
    xx = purge_crs_info(xx_epsg4326)
    assert xx.odc.crs is None
    xx = xx.assign_attrs(crs="epsg:4326")
    xx.latitude.attrs["crs"] = epsg3577
    with pytest.warns(UserWarning):
        _crs = xx.odc.crs
        assert _crs is not None
        assert _crs in (epsg3577, "epsg:4326")


def test_geobox_hook(xx_epsg4326: xr.DataArray):
    register_geobox()
    xx = xx_epsg4326
    assert xx.odc.crs == "epsg:4326"
    assert xx.geobox == xx.odc.geobox
    assert xx.to_dataset(name="xx").geobox == xx.odc.geobox

    assert purge_crs_info(xx)[:1, :1].geobox is None
    assert purge_crs_info(xx)[:1, :1].to_dataset(name="xx").geobox is None


def test_wrap_xr():
    gbox = GeoBox.from_bbox([0, -10, 100, 20], "epsg:4326", tight=True, shape=(13, 29))
    data = np.zeros(gbox.shape, dtype="uint16")

    xx = wrap_xr(data, gbox)
    assert xx.shape == gbox.shape
    assert xx.odc.geobox == gbox
    assert xx.dims == gbox.dims
    assert xx.attrs == {}

    assert wrap_xr(data, gbox, always_yx=True).dims == ("y", "x")
    assert wrap_xr(data, gbox, dims=("Y", "X")).dims == ("Y", "X")

    xx = wrap_xr(data, gbox)
    assert xx.attrs == {}

    xx = wrap_xr(data, gbox, nodata=10, some_flag=3)
    assert xx.attrs == dict(nodata=10, some_flag=3)

    xx = wrap_xr(data, gbox, time="2022-02-02T22:22:22.222222")
    assert xx.time.dt.year.item() == 2022
    assert xx.time.dt.month.item() == 2

    xx = wrap_xr(data[np.newaxis, ...], gbox, time="2022-02-02T22:22:22.222222")
    assert xx.shape == (1, *gbox.shape)
    assert xx.time.dt.year.values[0] == 2022
    assert xx.time.dt.month.values[0] == 2

    xx = wrap_xr(data[np.newaxis, ...], gbox, time=["2022-02-02T22:22:22.222222"])
    assert xx.shape == (1, *gbox.shape)
    assert xx.time.dt.year.values[0] == 2022
    assert xx.time.dt.month.values[0] == 2

    xx = wrap_xr(data[..., np.newaxis], gbox)
    assert xx.shape == (*gbox.shape, 1)
    assert xx.band.data.tolist() == ["b0"]

    assert wrap_xr(data.astype("float32"), gbox).odc.nodata is None
    assert np.isnan(wrap_xr(data.astype("float32"), gbox, nodata=NaN).odc.nodata)


@pytest.mark.parametrize("gbox", TEST_GEOBOXES_SMALL_AXIS_ALIGNED)
@pytest.mark.parametrize("nprefix", [0, 1, 2])
@pytest.mark.parametrize("npostfix", [0, 1, 2])
def test_wrap_xr_nd(gbox: GeoBox, nprefix: int, npostfix: int):
    shape = (1,) * nprefix + gbox.shape + (3,) * npostfix
    data = np.zeros(shape, dtype="uint16")
    xx = wrap_xr(data, gbox, axis=nprefix)
    assert xx.odc.geobox == gbox
    assert xx.odc.ydim == nprefix
    assert xx.dims[:nprefix] == ("time", "dim_0", "dim_1", "dim_2", "dim_3")[:nprefix]

    if npostfix == 1:
        assert xx.dims[-1] == "band"
    if npostfix > 1:
        assert xx.dims[nprefix + 2 :] == tuple(f"b_{i}" for i in range(npostfix))

    _dims = tuple(f"custom_{dim}" for dim in xx.dims)
    _dims = _dims[:nprefix] + ("y", "x") + _dims[nprefix + 2 :]

    yy = wrap_xr(data, gbox, axis=nprefix, dims=_dims)
    assert yy.dims == _dims
    assert yy.odc.geobox == gbox
    assert yy.odc.spatial_dims == _dims[nprefix : nprefix + 2]


@pytest.mark.parametrize("xx_time", [None, ["2020-01-30"]])
@pytest.mark.parametrize("xx_chunks", [None, (-1, -1), (4, 4)])
def test_xr_reproject(xx_epsg4326: xr.DataArray):
    assert isinstance(xx_epsg4326.odc, ODCExtensionDa)
    xx0 = xx_epsg4326
    xx0.attrs["crs"] = "epsg:4326"
    # smoke-test only
    assert isinstance(xx_epsg4326.odc.geobox, GeoBox)
    assert xx_epsg4326.odc.nodata is None
    dst_gbox = xx_epsg4326.odc.geobox.zoom_out(1.3)
    xx = xx_epsg4326.odc.reproject(dst_gbox)
    assert isinstance(xx, xr.DataArray)
    assert isinstance(xx.odc, ODCExtensionDa)

    assert xx.odc.geobox == dst_gbox
    assert xx.encoding["grid_mapping"] == "spatial_ref"
    assert "crs" not in xx.attrs
    assert xx.dtype == xx_epsg4326.dtype

    assert xx_epsg4326.odc.reproject(3857, dtype="float32").dtype == "float32"

    yy = xr.Dataset({"a": xx0, "b": xx0 + 1, "c": xr.DataArray([2, 3, 4])})
    assert isinstance(yy.odc, ODCExtensionDs)
    assert yy.odc.geobox == xx0.odc.geobox
    yy_ = yy.odc.reproject(dst_gbox)
    assert isinstance(yy_.odc, ODCExtensionDs)
    assert yy_.odc.geobox == dst_gbox
    assert yy_.a.odc.geobox == dst_gbox
    assert yy_.b.odc.geobox == dst_gbox
    assert (yy_.c == yy.c).all()

    yy_ = yy.odc.reproject("utm")
    assert yy_.odc.geobox.crs.proj.utm_zone is not None

    yy_ = xr_reproject(yy, "utm-n")
    assert isinstance(yy_, xr.Dataset)
    assert yy_.odc.geobox.crs.proj.utm_zone is not None
    assert yy_.odc.geobox.crs.proj.utm_zone.endswith("N")

    xx = xr_reproject(xx0, "utm-s")
    assert isinstance(xx, xr.DataArray)
    assert xx.odc.geobox.crs.proj.utm_zone is not None
    assert xx.odc.geobox.crs.proj.utm_zone.endswith("S")

    # check crs input
    xx = xx_epsg4326.odc.reproject("epsg:3857")
    assert xx.odc.geobox.crs == "epsg:3857"
    assert xx.odc.geobox.shape == xx_epsg4326.odc.output_geobox("epsg:3857").shape
    assert xx.odc.nodata is None

    # check dst_nodata override
    xx = xx_epsg4326.odc.reproject("epsg:3857", dst_nodata=255)
    assert xx.odc.nodata == 255

    # multi-time should work just the same
    if "time" in xx_epsg4326.dims:
        assert (xx_epsg4326.time == xx.time).all()
    else:
        xx2 = xx_epsg4326.expand_dims(time=2).odc.reproject("epsg:3857", dst_nodata=255)
        assert xx2.shape[0] == 2
        np.testing.assert_array_equal(xx2[0], xx)
        np.testing.assert_array_equal(xx2[1], xx)

    xx_i8 = xx_epsg4326.astype("int8")
    assert xx_i8.odc.reproject("epsg:3857").dtype == "int8"

    # non-georegistered case
    with pytest.raises(ValueError):
        _ = purge_crs_info(xx_epsg4326).odc.reproject(dst_gbox)

    with pytest.raises(ValueError):
        _ = xr.Dataset().odc.reproject("utm")


def test_xr_rasterize():
    gg = ocean_geom()
    xx = rasterize(gg, 1)
    assert xx.odc.geobox.crs == gg.crs
    yy = rasterize(gg, xx.odc.geobox)
    assert (xx == yy).all()

    yy = rasterize(gg, xx.odc.geobox[3:-4, 4:-8])
    assert (xx[3:-4, 4:-8] == yy).all()

    # chunk of ocean just off Africa in 3857
    xx = rasterize(gg, GeoBox.from_bbox([-20, -10, 20, 10], "epsg:3857", resolution=1))
    assert isinstance(xx.odc, ODCExtensionDa)
    assert xx.odc.geobox is not None
    assert xx.odc.geobox.crs == "epsg:3857"
    assert xx.all().item() is True

    # same but inverse
    xx = rasterize(
        gg,
        GeoBox.from_bbox([-20, -10, 20, 10], "epsg:3857", resolution=1),
        value_inside=False,
    )
    assert xx.odc.geobox.crs == "epsg:3857"
    assert xx.any().item() is False


def test_is_dask_collection():
    import dask

    import odc.geo._interop

    assert "is_dask_collection" in dir(odc.geo._interop)
    assert is_dask_collection is dask.is_dask_collection


@pytest.mark.parametrize("geobox", TEST_GEOBOXES_SMALL_AXIS_ALIGNED)
@pytest.mark.parametrize(
    "bad_geo_transform",
    [
        "some random string",
        "1 2 3 4 5 not-a-float",
    ],
)
def test_extract_transform(geobox, bad_geo_transform: str):
    xx = xr_zeros(geobox, dtype="int16")
    assert xx.odc.geobox == geobox
    assert _extract_geo_transform(xx.spatial_ref) == geobox.affine

    bad_attr_coord = xx.spatial_ref.copy()
    bad_attr_coord.attrs["GeoTransform"] = bad_geo_transform
    assert _extract_geo_transform(bad_attr_coord) is None


@pytest.mark.parametrize("geobox", TEST_GEOBOXES_SMALL_AXIS_ALIGNED)
def test_reload(geobox):
    xx = xr_zeros(geobox, dtype="int16")
    assert xx.odc.geobox == geobox
    assert xx.odc.reload() is xx
    assert xx.odc.geobox == geobox

    ds = xx.to_dataset(name="test")
    assert ds.odc.geobox == geobox
    assert ds.odc.reload() is ds
    assert ds.odc.geobox == geobox

    # change coords in-place
    coord = xx[list(xx.coords)[0]]
    coord.data[:] += 0.337
    assert xx.odc.reload() is xx
    assert xx.odc.geobox != geobox


@pytest.mark.parametrize("geobox", TEST_GEOBOXES_SMALL_AXIS_ALIGNED)
@pytest.mark.parametrize(
    "roi",
    [
        np.s_[:1, :1],
        np.s_[-1:, -1:],
        np.s_[1:2, 3:4],
        np.s_[1:2, :5],
        np.s_[1:, 3:4],
    ],
)
def test_geobox_1px(geobox: GeoBox, roi: ROI):
    assert geobox.axis_aligned
    assert min(geobox.shape) > 1
    assert max(geobox.shape) < 1000

    xx = xr_zeros(geobox, "uint8")
    assert isinstance(xx.odc, ODCExtensionDa)
    assert xx.odc.geobox == geobox

    assert approx_equal_geobox(xx[roi].odc.geobox, geobox[roi])

    # Verify that missing GeoTransform is handled without exceptions
    yy = xx.copy()
    yy.spatial_ref.attrs.pop("GeoTransform")
    assert min(yy[roi].shape[-2:]) == 1
    assert yy[roi].odc.geobox is None


@pytest.mark.parametrize("geobox", TEST_GEOBOXES_SMALL_AXIS_ALIGNED)
@pytest.mark.parametrize(
    "roi",
    [
        np.s_[1:1, :1],
        np.s_[-1:, -1:1],
        np.s_[1:1, 1:1],
    ],
)
def test_geobox_0px(geobox: GeoBox, roi: ROI):
    assert geobox.axis_aligned
    assert min(geobox.shape) > 1
    assert max(geobox.shape) < 1000

    xx = xr_zeros(geobox, "uint8")
    assert isinstance(xx.odc, ODCExtensionDa)
    assert xx.odc.geobox == geobox

    assert xx[roi].odc.geobox is None


@pytest.mark.parametrize(
    "poly, expected_fail",
    [
        (
            geom.polygon(
                ((-8, 8), (6, 1), (-4, -6)),
                crs="EPSG:4326",
            ),
            False,  # Fully inside, matching CRS
        ),
        (
            geom.polygon(
                ((-24, 8), (-10, 1), (-20, -6)),
                crs="EPSG:4326",
            ),
            False,  # Overlapping, matching CRS
        ),
        (
            geom.polygon(
                ((-40, 8), (-26, 1), (-36, -6)),
                crs="EPSG:4326",
            ),
            True,  # Fully outside, matching CRS
        ),
        (
            geom.polygon(
                ((-890555, 893463), (667916, 111325), (-445277, -669141)),
                crs="EPSG:3857",
            ),
            False,  # Fully inside, different CRS
        ),
        (
            geom.polygon(
                ((-2671667, 893463), (-1113194, 111325), (-2226389, -669141)),
                crs="EPSG:3857",
            ),
            False,  # Overlapping, different CRS
        ),
        (
            geom.polygon(
                ((-4452779, 893463), (-2894306, 111325), (-4007501, -669141)),
                crs="EPSG:3857",
            ),
            True,  # Fully outside, different CRS
        ),
    ],
)
def test_crop(xx_epsg4326, poly, expected_fail):
    xx = xx_epsg4326

    # If fail is expected, pass test
    if expected_fail:
        with pytest.raises(ValueError):
            xx_cropped = xx.odc.crop(poly=poly)
        return

    # Crop with default settings
    xx_cropped = xx.odc.crop(poly=poly)

    # Verify that cropped data is smaller
    assert xx_cropped.size <= xx.size

    # Verify that resolution and alignment have not changed
    np.testing.assert_array_almost_equal(
        xx.odc.geobox.resolution.xy, xx_cropped.odc.geobox.resolution.xy
    )
    np.testing.assert_array_almost_equal(
        xx.odc.geobox.alignment.xy, xx_cropped.odc.geobox.alignment.xy
    )

    # Verify that data contains NaN from default masking step
    assert xx_cropped.isnull().any()

    # Verify that no NaNs exist if masking not applied
    xx_nomask = xx.odc.crop(poly=poly, apply_mask=False)
    assert xx_nomask.notnull().all()

    # Verify that cropping also works on datasets
    xx_ds = xx.to_dataset(name="test")
    xx_ds_cropped = xx_ds.odc.crop(poly=poly)
    assert xx_ds_cropped.test.size <= xx_ds.test.size
    np.testing.assert_array_almost_equal(
        xx_ds.odc.geobox.resolution.xy, xx_ds_cropped.odc.geobox.resolution.xy
    )
    np.testing.assert_array_almost_equal(
        xx_ds.odc.geobox.alignment.xy, xx_ds_cropped.odc.geobox.alignment.xy
    )


@pytest.mark.parametrize(
    "poly",
    [
        geom.polygon(((-8, 8), (6, 1), (-4, -6)), crs="EPSG:4326"),
        geom.polygon(
            ((-890555, 893463), (667916, 111325), (-445277, -669141)), crs="EPSG:3857"
        ),
    ],
)
def test_mask(xx_epsg4326, poly):
    # Create test data and replace values with random integers so we can
    # reliably test that pixel values are the same before and after masking
    xx = xx_epsg4326
    xx.values[:] = np.random.randint(0, 10, size=xx.shape)

    # Apply mask
    xx_masked = xx.odc.mask(poly)

    # Verify that geobox is the same
    assert xx_masked.odc.geobox == xx.odc.geobox

    # Verify that data contains NaN from default masking step
    assert xx_masked.isnull().any()

    # Verify that non-masked values are the same as in non-masked dataset
    masked_pixels = np.isfinite(xx_masked.data)
    np.testing.assert_array_almost_equal(
        xx_masked.data[masked_pixels], xx.data[masked_pixels]
    )

    # Verify that `all_touched=False` produces fewer unmasked pixels
    xx_notouched = xx.odc.mask(poly, all_touched=False)
    assert xx_notouched.notnull().sum() < xx_masked.notnull().sum()

    # Verify that masking also works on datasets
    xx_ds = xx.to_dataset(name="test")
    xx_ds_masked = xx_ds.odc.mask(poly=poly)
    assert xx_ds_masked.odc.geobox == xx_ds.odc.geobox
    assert xx_ds_masked.test.isnull().any()


def test_spatial_dims():
    gbox = GeoBox.from_bbox([0, -10, 100, 20], "epsg:4326", tight=True, shape=(13, 29))
    xx0 = xr_zeros(gbox, "int16", always_yx=True, time=["2020-01-01", "2020-01-02"])
    assert spatial_dims(xx0) == ("y", "x")

    assert wrap_xr(
        xx0.data,
        gbox,
        dims=("T", "Y", "X"),
        axis=1,
    ).odc.spatial_dims == ("Y", "X")
    assert wrap_xr(
        xx0.data[..., np.newaxis],
        gbox,
        dims=("time", "Y", "X", "wavelength"),
        axis=1,
    ).odc.spatial_dims == ("Y", "X")

    xx = wrap_xr(
        np.zeros((2, 3, *gbox.shape.yx, 2, 3), dtype="int16"),
        gbox,
        axis=2,
        dims=("A", "B", "Y", "X", "C", "D"),
    )
    assert "X" in xx.coords and "Y" in xx.coords
    assert xx.odc.geobox == gbox
    assert xx.odc.ydim == 2
    assert xx.odc.spatial_dims == xx.dims[2:4]



from functools import partial
from typing import Any, Optional, Sequence, Tuple, Union, Mapping
from uuid import uuid4

import dask.array as da
import numpy as np
from dask.highlevelgraph import HighLevelGraph

from ._blocks import BlockAssembler
from .gcp import GCPGeoBox
from .geobox import GeoBox, GeoboxTiles
from .math import resolve_fill_value
from .types import Nodata
from .warp import Resampling, _rio_reproject, resampling_s2rio


def _do_chunked_reproject(
    d2s: Mapping[Tuple[int, int], Sequence[Tuple[int, int]]],
    src_gbt: GeoboxTiles,
    dst_gbt: GeoboxTiles,
    dst_idx: Tuple[int, int],
    *blocks: np.ndarray,
    axis: int = 0,
    dtype=None,
    casting="same_kind",
    resampling: Resampling = "nearest",
    src_nodata: Nodata = None,
    dst_nodata: Nodata = None,
    **kwargs,
):
    # pylint: disable=too-many-locals
    src_gbt, src_idx = src_gbt.clip(d2s[dst_idx])
    src_gbox = src_gbt.base
    dst_gbox = dst_gbt[dst_idx]
    assert isinstance(dst_gbox, GeoBox)
    assert isinstance(src_gbox, (GCPGeoBox, GeoBox))

    ba = BlockAssembler(dict(zip(src_idx, blocks)), src_gbt.chunks, axis=axis)
    if dtype is None:
        dtype = ba.dtype

    dst_shape = ba.with_yx(ba.shape, dst_gbox.shape)
    dst = np.full(
        dst_shape,
        resolve_fill_value(dst_nodata, src_nodata, dtype),
        dtype=dtype,
    )

    for src_roi in ba.planes_yx():
        src = ba.extract(src_nodata, dtype=dtype, casting=casting, roi=src_roi)
        dst_roi = ba.with_yx(src_roi, np.s_[:, :])

        _ = _rio_reproject(
            src,
            dst[dst_roi],
            src_gbox,
            dst_gbox,
            resampling=resampling,
            src_nodata=src_nodata,
            dst_nodata=dst_nodata,
            **kwargs,
        )

    return dst


def dask_rio_reproject(
    src: da.Array,
    s_gbox: Union[GeoBox, GCPGeoBox],
    d_gbox: GeoBox,
    resampling: Resampling,
    src_nodata: Nodata = None,
    dst_nodata: Nodata = None,
    ydim: int = 0,
    chunks: Optional[Tuple[int, int]] = None,
    dtype=None,
    **kwargs,
) -> da.Array:
    # pylint: disable=too-many-arguments, too-many-locals
    if isinstance(resampling, str):
        resampling = resampling_s2rio(resampling)

    if chunks is None:
        ny, nx = map(int, src.chunksize[ydim : ydim + 2])
        chunks = (ny, nx)

    def with_yx(a, yx):
        return (*a[:ydim], *yx, *a[ydim + 2 :])

    if dtype is None:
        dtype = src.dtype

    name: str = kwargs.pop("name", "reproject")

    gbt_src = GeoboxTiles(s_gbox, src.chunks[ydim : ydim + 2])
    gbt_dst = GeoboxTiles(d_gbox, chunks)
    d2s_idx = gbt_dst.grid_intersect(gbt_src)

    dst_shape = with_yx(src.shape, d_gbox.shape.yx)
    dst_chunks: Tuple[Tuple[int, ...], ...] = with_yx(src.chunks, gbt_dst.chunks)
    fill_value = resolve_fill_value(dst_nodata, src_nodata, dtype)

    tk = uuid4().hex
    name = f"{name}-{tk}"
    dsk: Any = {}

    proc = partial(
        _do_chunked_reproject,
        d2s_idx,
        gbt_src,
        gbt_dst,
        src_nodata=src_nodata,
        dst_nodata=fill_value,
        axis=ydim,
        resampling=resampling,
        dtype=dtype,
        **kwargs,
    )
    src_block_keys = src.__dask_keys__()

    def _src(idx):
        a = src_block_keys
        for i in idx:
            a = a[i]
        return a

    shape_in_blocks = tuple(map(len, dst_chunks))
    for idx in np.ndindex(shape_in_blocks):
        y, x = idx[ydim : ydim + 2]
        srcs = [with_yx(idx, (y, x)) for y, x in d2s_idx.get((y, x), [])]

        k = (name, *idx)
        if srcs:
            block_deps = tuple(_src(s_idx) for s_idx in srcs)
            dsk[k] = (proc, (y, x), *block_deps)
        else:
            b_shape = tuple(ch[i] for ch, i in zip(dst_chunks, idx))
            dsk[k] = (np.full, b_shape, fill_value, src.dtype)

    dsk = HighLevelGraph.from_collections(name, dsk, dependencies=(src,))

    return da.Array(dsk, name, chunks=dst_chunks, dtype=dtype, shape=dst_shape)



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2022 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
from typing import List, Optional, Tuple, Union

import numpy as np
from affine import Affine

from .crs import CRS, MaybeCRS, SomeCRS, norm_crs
from .geobox import GeoBox, GeoBoxBase
from .geom import Geometry, multipoint
from .math import Poly2d, affine_from_pts, align_up, resolution_from_affine, unstack_xy
from .types import XY, MaybeInt, Resolution, SomeResolution, SomeShape, wh_

SomePointSet = Union[np.ndarray, Geometry, List[Geometry], List[XY[float]]]


def _points_to_array(pts: SomePointSet) -> Tuple[np.ndarray, Optional[CRS]]:
    if isinstance(pts, np.ndarray):
        return pts, None
    if isinstance(pts, Geometry):
        return np.asarray([pt.coords[0] for pt in pts.geoms]), pts.crs

    def _xy(pt: Union[Geometry, XY[float]]) -> Tuple[float, float]:
        if isinstance(pt, Geometry):
            ((x, y),) = pt.coords
        else:
            x, y = pt.xy
        return (x, y)

    crs = getattr(pts[0], "crs", None)
    return np.asarray([_xy(pt) for pt in pts]), crs


class GCPMapping:
    """
    Shared state for GCPGeoBox.
    """

    def __init__(
        self,
        pix: SomePointSet,
        wld: SomePointSet,
        crs: MaybeCRS = None,
    ):
        pix, _ = _points_to_array(pix)
        wld, _crs = _points_to_array(wld)

        if crs is None:
            crs = _crs

        # Nx2
        assert pix.shape == wld.shape
        assert pix.shape[1] == 2

        self._pix = pix
        self._wld = wld
        self._crs = norm_crs(crs)
        self._approx_affine: Optional[Affine] = None

        self._p2w: Optional[Poly2d] = None
        self._w2p: Optional[Poly2d] = None

    @property
    def p2w(self) -> Poly2d:
        """Pixel to world."""
        if self._p2w is not None:
            return self._p2w

        self._p2w = Poly2d.fit(self._pix, self._wld)
        return self._p2w

    @property
    def w2p(self) -> Poly2d:
        """World to pixel."""
        if self._w2p is not None:
            return self._w2p

        self._w2p = Poly2d.fit(self._wld, self._pix)
        return self._w2p

    @property
    def crs(self) -> Optional[CRS]:
        return self._crs

    @property
    def approx(self) -> Affine:
        if self._approx_affine is not None:
            return self._approx_affine

        pix = unstack_xy(self._pix)
        wld = unstack_xy(self._wld)
        self._approx_affine = affine_from_pts(pix, wld)

        return self._approx_affine

    @property
    def resolution(self) -> Resolution:
        """Resolution, pixel size in CRS units."""
        return resolution_from_affine(self.approx)

    def points(self) -> Tuple[Geometry, Geometry]:
        """Return multipoint geometries for (Pixel, World)."""
        pix_points: list[tuple[float, float]] = [
            (float(p[0]), float(p[1])) for p in self._pix.tolist()
        ]
        wld_points: list[tuple[float, float]] = [
            (float(p[0]), float(p[1])) for p in self._wld.tolist()
        ]

        return (
            multipoint(pix_points, None),
            multipoint(wld_points, self.crs),
        )

    def __dask_tokenize__(self):
        return (
            "odc.geo._gcp.GCPMapping",
            str(self._crs),
            self._wld,
            self._pix,
        )

    @staticmethod
    def from_rio(src, output_crs: MaybeCRS = None) -> "GCPMapping":
        """
        Construct from an open rasterio file.
        """
        # pylint: disable=import-outside-toplevel
        from .converters import extract_gcps

        return GCPMapping(*extract_gcps(src, output_crs))


class GCPGeoBox(GeoBoxBase):
    """
    Ground Control Point based GeoBox.

    """

    __slots__ = ("_mapping",)

    def __init__(
        self, shape: SomeShape, mapping: GCPMapping, affine: Optional[Affine] = None
    ):
        if affine is None:
            affine = Affine.identity()
        GeoBoxBase.__init__(self, shape, affine, mapping.crs)
        self._mapping = mapping

    def __getitem__(self, roi) -> "GCPGeoBox":
        _shape, _affine = self.compute_crop(roi)
        return GCPGeoBox(shape=_shape, mapping=self._mapping, affine=_affine)

    @property
    def approx(self) -> GeoBox:
        """
        Compute best fitting linear GeoBox.
        """
        return GeoBox(self.shape, self._mapping.approx * self._affine, self.crs)

    @property
    def resolution(self) -> Resolution:
        """Resolution, pixel size in CRS units."""
        return self.approx.resolution

    def wld2pix(self, x, y):
        x, y = self._mapping.w2p(x, y)
        return (~self._affine) * (x, y)

    def pix2wld(self, x, y):
        x, y = self._affine * (x, y)
        wx, wy = self._mapping.p2w(x, y)
        return (wx, wy)

    def __hash__(self):
        return hash((*self._shape, self._affine, self._crs, id(self._mapping)))

    @property
    def linear(self) -> bool:
        return False

    @property
    def axis_aligned(self):
        return False

    @property
    def center_pixel(self) -> "GCPGeoBox":
        """
        GeoBox of a center pixel.
        """
        return self[self.shape.map(lambda x: x // 2).yx]

    def map_bounds(self) -> Tuple[Tuple[float, float], Tuple[float, float]]:
        """
        Query bounds in folium/ipyleaflet style.

        Returns SW, and NE corners in lat/lon order.
        ``((lat_w, lon_s), (lat_e, lon_n))``.
        """
        if self._crs is not None:
            x0, y0, x1, y1 = self.geographic_extent.boundingbox.bbox
        else:
            x0, y0, x1, y1 = self.extent.boundingbox.bbox
        return (y0, x0), (y1, x1)

    def to_crs(self, crs: SomeCRS, **_) -> "GCPGeoBox":
        """
        Project GCPs to a different CRS.

        extra arguments undesrtood by GeoBox.to_crs are simply ignored, but we do
        accept them to make generic code easier to write.
        """
        assert self._crs is not None
        pix, wld = self._mapping.points()
        if not self._affine.is_identity:
            pix = pix.transform(~self._affine)
        wld = wld.to_crs(crs)
        mapping = GCPMapping(pix, wld, wld.crs)
        return GCPGeoBox(self._shape, mapping)

    def pad(self, padx: int, pady: MaybeInt = None) -> "GCPGeoBox":
        """
        Pad geobox.

        Expand GeoBox by fixed number of pixels on each side
        """
        # false positive for -pady, it's never None by the time it runs
        # pylint: disable=invalid-unary-operand-type

        pady = padx if pady is None else pady

        ny, nx = self._shape.yx
        A = self._affine * Affine.translation(-padx, -pady)
        shape = (ny + pady * 2, nx + padx * 2)
        return GCPGeoBox(shape, self._mapping, A)

    def pad_wh(self, alignx: int = 16, aligny: MaybeInt = None) -> "GCPGeoBox":
        """
        Possibly expand :py:class:`~odc.geo.geobox.GeoBox` by a few pixels.

        Find nearest ``width``/``height`` that are multiples of the desired factor. And return a new
        geobox that is slighly taller and/or wider covering roughly the same region. The new geobox
        will have the same CRS and transform but possibly larger shape.
        """
        aligny = alignx if aligny is None else aligny
        ny, nx = (align_up(sz, n) for sz, n in zip(self._shape.yx, (aligny, alignx)))

        return GCPGeoBox((ny, nx), self._mapping, self._affine)

    def zoom_out(self, factor: float) -> "GCPGeoBox":
        """
        Compute :py:class:`~odc.geo.geobox.GCPGeoBox` with changed resolution.

        - ``factor > 1`` implies smaller width/height, fewer but bigger pixels
        - ``factor < 1`` implies bigger width/height, more but smaller pixels

        :returns:
           GCPGeoBox covering the same region but with different pixels (i.e. lower or higher resolution)
        """
        _shape, _affine = self.compute_zoom_out(factor)
        return GCPGeoBox(_shape, self._mapping, _affine)

    def zoom_to(
        self,
        shape: Union[SomeShape, int, float, None] = None,
        *,
        resolution: Optional[SomeResolution] = None,
    ) -> "GCPGeoBox":
        """
        Compute :py:class:`~odc.geo.geobox.GCPGeoBox` with changed resolution.

        When supplied a single integer scale longest dimension to match that.

        :returns:
          GCPGeoBox covering the same region but with different number of pixels and therefore resolution.
        """
        _shape, _affine = self.compute_zoom_to(shape, resolution=resolution)
        return GCPGeoBox(_shape, self._mapping, _affine)

    def __str__(self):
        return self.__repr__()

    def __repr__(self) -> str:
        return f"GCPGeoBox({self._shape.xy!r}, {self.crs!r})"

    def __eq__(self, __o: object) -> bool:
        if not isinstance(__o, GCPGeoBox):
            return False

        return (
            self._shape == __o.shape
            and self._mapping is __o._mapping
            and self._affine == __o._affine
        )

    def gcps(self):
        """Extract GCPs in rasterio format."""
        # pylint: disable=import-outside-toplevel
        from rasterio.control import GroundControlPoint

        pix, wld = self._mapping.points()

        a_inv = ~self._affine

        def to_gcp(pix: Geometry, wld: Geometry, _id) -> GroundControlPoint:
            ((wx, wy),) = wld.coords

            ((px, py),) = pix.coords
            px, py = a_inv * (px, py)
            return GroundControlPoint(row=py, col=px, x=wx, y=wy, id=_id)

        return [
            to_gcp(p, w, idx) for idx, (p, w) in enumerate(zip(pix.geoms, wld.geoms))
        ]

    def __dask_tokenize__(self):
        return (
            "odc.geo._gcp.GCPGeoBox",
            *self._mapping.__dask_tokenize__()[1:],
            *self._shape.yx,
            *self._affine[:6],
        )

    @staticmethod
    def from_rio(src, output_crs: MaybeCRS = None) -> "GCPGeoBox":
        """
        Construct from an open rasterio file.
        """
        return GCPGeoBox(
            wh_(src.width, src.height), GCPMapping.from_rio(src, output_crs)
        )



from odc.geo.geobox import GeoBox
from odc.geo.ui import make_svg, svg_base_map


def test_svg_base_smoke_only():
    gbox = GeoBox.from_bbox((110, -45, 160, -10), resolution=0.05)
    assert "<svg" in svg_base_map()

    assert "</svg>" in svg_base_map(gbox, gbox.extent, "<path/>")

    # long
    assert "</svg>" in svg_base_map("<path/>", bbox=(-10, 0, 10, 1))

    # tall
    assert "</svg>" in svg_base_map("<path/>", bbox=(0, 0, 1, 10))

    # cross-hair
    assert "</svg>" in svg_base_map(target=(0, 10))


def test_make_svg_smoke_only():
    gbox = GeoBox.from_bbox((110, -45, 160, -10), resolution=0.05)

    assert "</svg>" in make_svg()
    assert "</svg>" in make_svg(gbox)
    # long
    assert "</svg>" in make_svg("<path/>", bbox=(-10, 0, 10, 1))

    # tall
    assert "</svg>" in make_svg("<path/>", bbox=(0, 0, 1, 10))
